#                       Copyright 2021 Saad Mairaj
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import colour
import tkinter
from tkinter.font import Font
from tkmacosx.utils import (_cnfmerge, _bind, _Canvas, check_param,
                            check_light_dark, delta, gradient)

HEX = 'hex'
RGB = 'rgb'

COLORSCALE_PROPERTIES = [
    'borderwidth', 'cursor', 'command', 'gradient', 'height',
    'highlightbackground', 'highlightcolor', 'highlightthickness',
    'mousewheel', 'orient', 'relief', 'showinfo', 'showinfodelay',
    'state', 'takefocus', 'value', 'variable', 'width'
]

COLORSCALE_FEATURES = (
    'value', 'command', 'orient', 'mousewheel', 'variable', 'showinfo',
    'showinfodelay', 'gradient'
)


class ColorscaleBase(_Canvas):
    "Base class for Colorscale widget."

    def __init__(self, master=None, cnf={}, **kw):
        kw = {k: v for k, v in _cnfmerge(
            (cnf, kw)).items() if v is not None}
        cnf = {k: kw.pop(k, None)
               for k in kw.copy() if k in COLORSCALE_FEATURES}

        cnf['command'] = cnf.get('command', '')
        cnf['gradient'] = cnf.get('gradient', 'default')
        cnf['mousewheel'] = cnf.get('mousewheel', False)
        cnf['orient'] = cnf.get('orient', 'vertical')
        cnf['showinfo'] = cnf.get('showinfo', True)
        cnf['showinfodelay'] = cnf.get('showinfodelay', 1500)
        cnf['value'] = cnf.get('value', 'hex')
        cnf['variable'] = cnf.get('variable', '')

        kw['width'] = kw.get(
            "width", 250 if 'ver' in cnf['orient'] else 30)
        kw['height'] = kw.get(
            "height", 30 if 'ver' in cnf['orient'] else 250)
        kw['highlightthickness'] = kw.get('highlightthickness', 0)
        _Canvas.__init__(self, master=master, cnf={}, **kw)

        self.cnf = self._check_param(**cnf)
        self._size = (0, 0)
        self._marker_color = 'black'
        self._xy = int((self.winfo_reqwidth() if 'ver' in
                        self.cnf['orient'] else self.winfo_reqheight()) / 3)

        binds = [
            {'className': 'set_size', 'sequence': '<Configure>', 'func': self._set_size},
            {'className': 'b1_on_motion', 'sequence': '<B1-Motion>',
                'func': self._check_marker},
            {'className': 'b1_press', 'sequence': '<Button-1>',
                'func': self._check_marker}
        ]
        _bind(self, *binds)
        self._set_mousewheel()

    def _check_param(self, cnf={}, **kw):
        kw = _cnfmerge((cnf, kw))
        if 'gradient' in kw:
            if kw['gradient'] != 'default':
                cond = isinstance(kw['gradient'], (list, tuple))
                if not cond or (cond and len(kw['gradient']) <= 1):
                    raise tkinter.TclError(
                        'expected sequence of two color values but got "%s"'
                        % kw['gradient'])
                for c in kw['gradient']:
                    self.winfo_rgb(c)
        kw = check_param(self, 'colorscale', kw)
        return kw

    def _set_size(self, evt=None):
        """Internal function."""
        if evt.width == self._size[0] and evt.height == self._size[1]:
            return
        self._size = (evt.width, evt.height)
        self._create_items('create', safe_create=True)

    def _callback(self, color):
        """Internal function."""
        if self.cnf.get('command'):
            self.cnf['command'](color)
        if self.cnf.get('variable'):
            self.cnf['variable'].set(color)

    def _create_items(self, cmd, safe_create=False, **kw):
        """Internal function.\n
        Checks and creates (text, marker, 
        showinfo, gradient lines) items."""

        def check_tag(tag):
            return self.check_tag(cmd, tag, safe_create, kw.get('avoid', []))

        ids = []

        if check_tag('gradient'):
            w, h = self.winfo_width(), self.winfo_height()
            iteration = w if 'ver' in self.cnf['orient'] else h
            color_list = gradient(iteration)
            if isinstance(self.cnf.get('gradient'), (list, tuple)):
                c1 = colour.Color(self.cnf.get('gradient')[0])
                c2 = colour.Color(self.cnf.get('gradient')[1])
                color_list = c1.range_to(c2, iteration)
            elif isinstance(self.cnf.get('gradient'), str) \
                    and self.cnf.get('gradient') != 'default':
                c = colour.Color(self.cnf.get('gradient'))
                color_list = c.range_to(c, iteration)

            for count, c in enumerate(color_list):
                ags = (count, -1, count, h)
                if self.cnf['orient'] == 'horizontal':
                    ags = (-1, count, w, count)
                ids.append(self._create('line', ags, {
                           'fill': c, 'tag': 'gradient'}))

        if check_tag('borderline'):
            borderline_points = kw.get('borderline_points',
                                       (1, 1, self.winfo_width()-2,
                                        self.winfo_height()-2, 0))
            ids.append(self.rounded_rect(borderline_points, width=2,
                                         outline='#81b3f4', tag='borderline',
                                         style='arc'))

        if check_tag('marker'):
            _def_val = (2, self._xy, self.winfo_width()-4, 5, 2)
            if self['orient'] == 'vertical':
                _def_val = (self._xy, 2, 5, self.winfo_height()-4, 2)
            marker_points = kw.get('marker_points', _def_val)
            ids.append(self.rounded_rect(marker_points, width=2,
                                         outline=self._marker_color,
                                         tag="marker", style='arc'))

        if check_tag('markerbg') and kw.get('markerbg_points') is not None:
            markerbg_points = kw.get('markerbg_points')
            cnf = kw.get('markerbg_cnf')
            ids.append(self._rounded(markerbg_points, **cnf))

        if check_tag('info') and kw.get('info_points') is not None:
            info_points = kw.get('info_points')
            cnf = kw.get('info_cnf')
            ids.append(self._create('text', info_points, cnf))

        return ids

    def _release(self, evt=None):
        """Internal function."""
        self.delete('info', 'markerbg')
    
    def _move_marker(self, evt):
        spacer, spacbg = 35, 25
        c_id = self.find('overlapping', evt.x, evt.y, evt.x, evt.y)
        text = hexcode = self.itemcget(c_id[0], 'fill')
        rgb = [int(i/65535.0*255.0) for i in self.winfo_rgb(hexcode)]
        self._marker_color = check_light_dark(rgb)

        self._configure(('itemconfigure', 'borderline'),
                        {'outline': hexcode}, None)
        self._callback(hexcode)

        if self.cnf['value'] == "rgb":
            spacer, spacbg = 65, 55
            text = ' | '.join([v+':'+str(f)
                               for f, v in zip(rgb, ('R', 'G', 'B'))])
            self._callback(rgb)

        ver_cond = evt.x < self.winfo_width() - (spacbg+spacer) \
            and self['orient'] == 'vertical'
        hor_cond = evt.y < self.winfo_height() - (spacbg+spacer) \
            and self['orient'] == 'horizontal'

        markerbg_points, info_points = (), ()

        if bool(ver_cond or hor_cond) and self['showinfo']:
            markerbg_points = (self.winfo_width()/2-6, evt.y+spacer - spacbg,
                               self.winfo_width()/2+7, evt.y+spacer + spacbg, 6)
            info_points = ((self.winfo_width()/2, evt.y+spacer))

            if self.cnf['orient'] == 'vertical':
                markerbg_points = (evt.x+spacer - spacbg, self.winfo_height()/2-6,
                                   evt.x+spacer + spacbg, self.winfo_height()/2+7, 6)
                info_points = ((evt.x + spacer, self.winfo_height()/2))

        elif self['showinfo']:
            markerbg_points = (self.winfo_width()/2-6, evt.y-spacer - spacbg,
                               self.winfo_width()/2+7, evt.y-spacer + spacbg, 6)
            info_points = ((self.winfo_width()/2, evt.y-spacer))

            if self.cnf['orient'] == 'vertical':
                markerbg_points = (evt.x-spacer - spacbg, self.winfo_height()/2-6,
                                   evt.x-spacer + spacbg, self.winfo_height()/2+7, 6)
                info_points = ((evt.x - spacer, self.winfo_height()/2))

        markerbg_cnf = {'fill': self._marker_color, 'tag': 'markerbg'}
        info_cnf = {'angle': 0 if 'ver' in self.cnf['orient'] else 90,
                    'text': text, 'font': Font(size=10), 'tag': "info", 'fill': hexcode}

        self._create_items('create', safe_create=True, avoid=('gradient', 'borderline'),
                           info_points=info_points, markerbg_points=markerbg_points,
                           info_cnf=info_cnf, markerbg_cnf=markerbg_cnf)
        return True

    def _check_marker(self, evt, mw=None):
        """Internal function."""
        diff_orient_val = {
            'vertical': dict(evtx=mw, evty=10, _xy=evt.x),
            'horizontal': dict(evtx=10, evty=mw, _xy=evt.y)}
        
        self.after_cancel(getattr(self, '_remove_id', ' '))
        self._remove_id = self.after(self.cnf['showinfodelay'], self._release)

        if mw:
            evt.x = diff_orient_val[self['orient']]['evtx']
            evt.y = diff_orient_val[self['orient']]['evty']
        else:
            self._xy = diff_orient_val[self['orient']]['_xy']

        cond_x = bool(evt.x > 0 and evt.x < self.winfo_width())
        cond_y = bool(evt.y > 0 and evt.y < self.winfo_height())
        cond_state = bool(self['state'] not in 'disabled')

        if not (cond_x and cond_y and cond_state):
            return
        return self._move_marker(evt)
        

    def _set_mousewheel(self, evt=None):
        """Internal function.\n
        Sets mousewheel scrolling."""

        def on_mousewheel(evt=None):
            "Internal function."
            ver_cond = self._xy < self.winfo_width() \
                and self['orient'] == 'vertical'
            hor_cond = self._xy < self.winfo_height() \
                and self['orient'] == 'horizontal'
            if delta(evt) <= -1 and (ver_cond or hor_cond):
                self._xy += 1
                if not self._check_marker(evt, mw=self._xy):
                    self._xy -= 1
            if delta(evt) >= 1 and self._xy > 1:
                self._xy -= 1
                if not self._check_marker(evt, mw=self._xy):
                    self._xy += 1

        if self.cnf.get('mousewheel'):
            _bind(self, className='mousewheel',
                  sequence='<MouseWheel>', func=on_mousewheel)
            _bind(self, className='mousewheel_x11',
                  sequence='<Button-4>', func=on_mousewheel)
            _bind(self, className='mousewheel_x11',
                  sequence='<Button-5>', func=on_mousewheel)
        else:
            _bind(self, className='mousewheel',
                  sequence='<MouseWheel>')
            _bind(self, className='mousewheel_x11',
                  sequence='<Button-4>')
            _bind(self, className='mousewheel_x11',
                  sequence='<Button-5>')

    def _configure(self, cmd, cnf=None, kw=None):
        """Internal function."""
        kw1 = _cnfmerge((cnf, kw))

        for i in list(kw1):
            if i in COLORSCALE_FEATURES:
                self.cnf.update(
                    self._check_param({i: kw1.pop(i)})
                )

        values = _Canvas._configure(self, cmd, None, kw1)
        if _cnfmerge((cnf, kw)).get('gradient'):
            self._create_items('create', safe_create=True,
                               avoid=('borderline', 'marker',
                                      'markerbg', 'info'))
        self._set_mousewheel()
        if values and isinstance(values, dict):
            values.update(self.cnf)
            for k in list(values):
                if k not in COLORSCALE_PROPERTIES:
                    values.pop(k)
        return values

    def cget(self, key):
        """Return the resource value for a KEY given as string."""
        if key in self.cnf.keys():
            if key == 'variable':
                return str(self.cnf[key])
            return self.cnf.get(key)
        value = _Canvas.cget(self, key)
        if key in ('height', 'width', 'highlightthickness',
                   'borderwidth', 'bd'):
            return int(value)
        return value
    __getitem__ = cget

    def keys(self):
        """Return a list of all resource names of this widget."""
        return COLORSCALE_PROPERTIES

    def destroy(self):
        """Destroy this widget."""
        self.after_cancel(getattr(self, '_remove_id', ' '))
        return _Canvas.destroy(self)



#                       Copyright 2021 Saad Mairaj
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import tkinter
from tkmacosx.utils import (SYSTEM_DEFAULT_BG, SYSTEM_DEFAULT_FG, 
                            _cnfmerge, _bind, _Canvas, check_param,
                            _info_button, _on_press_color, 
                            get_shade, check_function_equality)
import tkmacosx.utils.colorvar_patches as cp


BUTTON_PROPERTIES = [
    'activebackground', 'activeforeground', 'anchor', 'background', 
    'bd', 'bg', 'bitmap', 'borderwidth', 'command', 'compound', 'cursor', 
    'disabledforeground', 'fg', 'font', 'foreground', 'height', 
    'highlightbackground', 'highlightcolor', 'highlightthickness', 'image', 
    'justify', 'overrelief', 'padx', 'pady', 'relief', 'repeatdelay', 
    'repeatinterval', 'state', 'takefocus', 'text', 'textvariable', 
    'underline', 'width'
]


BUTTON_FEATURES = (
    'overbackground', 'overforeground', 'activeimage', 'activebitmap', 
    'anchor', 'bitmap', 'bordercolor', 'borderless', 'command', 'compound', 
    'disabledforeground', 'justify', 'disabledbackground', 'fg', 'font', 
    'foreground', 'height', 'image', 'overrelief', 'padx', 'pady', 'repeatdelay', 
    'repeatinterval', 'text', 'textvariable', 'underline', 'width', 'state', 
    'focusthickness', 'focuscolor', 'highlightbackground', 'activebackground', 
    'activeforeground'
)


BUTTON_ITEMS = (
    '_txt', '_bit', '_img', '_bd_color', '_border', '_tf'
)

BORDER_INTENSITY = 0.06

_warning_msg_shown = [False]


class _button_properties:
    """Internal class.\n
    Contains modified properties of Button widget. Do not call directly."""

    # Utils
    def _bit_img(self, cmd, kw):
        tag = '_bit' if cmd == 'bitmap' else '_img'
        if kw.get(cmd, '') != '':
            state = 'disabled' if self.cnf.get('state') in (
                'disabled', 'disable') else 'normal'
            _opt = dict(state=state, **{i: kw.get(i, self.cnf.get(i))
                                    for i in ('anchor', cmd)})
            return (('itemconfigure', tag), _opt, None)
        self.delete(tag)
    
    def _active_bit_img(self, cmd, kw):
        tag = '_bit' if cmd == 'activebitmap' else '_img'
        name = cmd.rsplit('active')[-1]
        if self['state'] in ('pressed', 'active'):
            return (('itemconfigure', tag), {
                    name: self.cnf.get(cmd, '')}, None)

    def _item(self, cmd, kw):
        self.cnf['width'] = self.winfo_width()
        self.cnf['height'] = self.winfo_height()
        
        r1 = self._compound(self.cnf.get('compound'), self.cnf.get('width'), self.cnf.get('height'))
        r2 = {cmd: ((self.cnf['width']/2), self.cnf['height']/2)}
        return r1 or r2
    
    def _size_opts(self):
        _opt = {}
        for i in ('text', 'font', 'textvariable', 'image', 'bitmap',
                    'compound', 'padx', 'pady', 'activeimage',
                    'activebitmap'):
            cond1 = self.cnf.get(i, '') != ''
            if cond1 and i == 'activeimage':
                _opt['image'] = self.cnf[i]
            elif cond1 and i == 'activebitmap':
                _opt['bitmap'] = self.cnf[i]
            elif cond1:
                _opt[i] = self.cnf.get(i)
        return _opt

    # These are properties of button widget:
    def _border(self, kw):
        return [('itemconfigure', '_border'),
            {'outline': get_shade(self['bg'], 
                BORDER_INTENSITY, 'auto-120')}, None]
    
    def _bit(self, kw):
        return _button_properties._item(self, '_bit', kw)

    def _img(self, kw):
        return _button_properties._item(self, '_img', kw)

    def _txt(self, kw):
        return _button_properties._item(self, '_txt', kw)

    def activebackground(self, kw):
        color = None
        if kw.get('activebackground', '') != '':
            color = kw.get('activebackground', self.cnf.get('activebackground'))
        return [self, {'tag': '_activebg',
                        'width':  self.winfo_width(),
                        'height': self.winfo_height(),
                        'color': color}]
    
    def activeforeground(self, kw):
        if self['state'] in ('pressed', 'active'):
            return (('itemconfigure', '_txt'), {
                    'fill': self.cnf.get('activeforeground', 'white')}, None)
    
    def activebitmap(self, kw):
        return _button_properties._active_bit_img(self, 'activebitmap', kw)
    
    def activeimage(self, kw):
        return _button_properties._active_bit_img(self, 'activeimage', kw)

    def bitmap(self, kw):
        return _button_properties._bit_img(self, 'bitmap', kw)

    def borderless(self, kw):
        def _get_master_bg():
            """Internal function."""
            try: 
                return self.master['bg']
            except tkinter.TclError:
                if not _warning_msg_shown[0]:
                    print('WARNING: "borderless" option of '
                        'tkmacosx Button doesn\'t work with ttk widgets. '
                        'Bordercolor/highlightbackground can be set manually '
                        'with "highlightbackground" or "bordercolor" options.\n')
                    _warning_msg_shown[0] = True
                return self.cnf.get('bordercolor', self.cnf['highlightbackground'])

        _opt = {}
        if bool(kw.get('borderless')) or self.cnf.get('borderless'):
            if not check_function_equality(self.master.config, self._get_functions('borderless', kw)):
                self.master.config = self.master.configure = self._get_functions(
                    'borderless', kw)
            self.cnf['highlightbackground'] = self.cnf['bordercolor'] = _get_master_bg()
            _opt[1] = [('itemconfigure', '_bd_color'), {'outline': _get_master_bg()}, None]
            _opt[2] = ['configure', {'highlightbackground': _get_master_bg()}, None]
        elif not bool(kw.get('borderless', True)) or not self.cnf.get('borderless'):
            if self.cnf.get('bordercolor') == _get_master_bg():
                self.cnf.pop('bordercolor', None)
                self.cnf.pop('highlightbackground', None)
            bd_color = self.cnf.get(
                'bordercolor', get_shade(self['bg'], 0.04, 'auto-120'))
            if bd_color == '':
                bd_color = get_shade(self['bg'], 0.04, 'auto-120')
            elif bd_color.lower() == 'default':
                bd_color = get_shade('white', 0.04, 'auto-120')
            self.cnf.update({'bordercolor': bd_color})
            self.cnf['highlightbackground'] = self.cnf['bordercolor'] = bd_color
            if self.itemcget('_bd_color', 'outline') != bd_color:
                _opt[1] = [('itemconfigure', '_bd_color'), {'outline': bd_color}, None]
                _opt[2] = ['configure', {'highlightbackground': bd_color}, None]
        return _opt
    bordercolor = highlightbackground = borderless

    def foreground(self, kw):
        return (('itemconfigure', '_txt'),
                {'fill': kw.get('foreground', self.cnf.get('foreground'))}, None)
    fg = foreground

    def focuscolor(self, kw):
        return (('itemconfigure', '_tf'),
                {'outline': kw.get('focuscolor', self.cnf.get('focuscolor'))}, None)
    
    def focusthickness(self, kw):
        _button_items.create_items(self, 'create', True, avoid=(
            '_txt', '_img', '_bit', '_bd_color', '_border'))
    
    def image(self, kw):
        return _button_properties._bit_img(self, 'image', kw)
    
    def overbackground(self, kw):
        _opt = []
        if kw.get('overbackground', '') != '':
            fn = self._get_functions('overbackground', kw)
            _opt = [self,
                    {'className': 'overbackground', 'sequence': '<Enter>',
                    'func': fn.get('<Enter>')},
                    {'className': 'overbackground', 'sequence': '<Leave>',
                    'func': fn.get('<Leave>')}]
            if self._mouse_state_condition():
                _Canvas._configure(self, 'configure', {'bg': kw.get('overbackground')}, None)
                _Canvas._configure(self, ('itemconfigure', '_border'), 
                    {'outline': get_shade(kw.get('overbackground'), 0.1, 'auto-120')}, None)
        elif kw.get('overbackground') == '':
            _opt = [self,
                    {'className': 'overbackground', 'sequence': '<Enter>'},
                    {'className': 'overbackground', 'sequence': '<Leave>'}, 
                    ('configure', {'bg': self._org_bg}, None), 
                    (('itemconfigure', '_border'), {'outline': 
                    get_shade(self._org_bg, 0.1, 'auto-120')}, None)]
        return _opt

    def overforeground(self, kw):
        _opt = []
        if kw.get('overforeground', '') != '':
            fn = self._get_functions('overforeground', kw)
            _opt = [self,
                    {'className': 'overforeground', 'sequence': '<Enter>',
                    'func': fn.get('<Enter>')},
                    {'className': 'overforeground', 'sequence': '<Leave>',
                    'func': fn.get('<Leave>')}]
            if self._mouse_state_condition():
                # [issue-3] doesn't change if overrelief is on.
                # [issue-3] (FIXED) using after with 0ms delay
                #           fixes the issue. To be safe delay is 1ms.
                self.after(1, lambda: _Canvas._configure(self,
                    ('itemconfigure', '_txt'), {'fill': kw.get('overforeground')}, None))
        elif kw.get('overforeground') == '':
            _opt = [self,
                    {'className': 'overforeground', 'sequence': '<Enter>'},
                    {'className': 'overforeground', 'sequence': '<Leave>'}]
            _opt.append((('itemconfigure', '_txt'),
                        {'fill': self.cnf.get('fg', 'black')}, None))
        return _opt

    def overrelief(self, kw):
        _opt = None
        if kw.get('overrelief', '') != '':
            if not self._rel[1]:
                self._rel = ('flat', False)
            if self._mouse_state_condition():
                self._configure(
                    'configure', {'_relief': kw.get('overrelief')}, None)
            fn = self._get_functions('overrelief', kw)
            _opt = (self,
                    {'className': 'overrelief', 'sequence': '<Enter>',
                        'func': fn.get('<Enter>')},
                    {'className': 'overrelief', 'sequence': '<Leave>',
                        'func': fn.get('<Leave>')})

        elif kw.get('overrelief') == '':
            _opt = (self,
                    {'className': 'overrelief', 'sequence': '<Enter>'},
                    {'className': 'overrelief', 'sequence': '<Leave>'},
                    ('configure', {'relief': self._rel[0]}, None))
        return _opt

    def size(self, kw):
        if self._type == 'circle' and kw.get('radius'):
            self.cnf['width'] = self.cnf['height'] = kw.get(
                'width', kw.get('height', int(kw['radius']*2)))
        if self._fixed_size['w'] and kw.get('width', True):
            kw['width'] = self.cnf['width']
        if self._fixed_size['h'] and kw.get('height', True):
            kw['height'] = self.cnf['height']
        self._fixed_size['w'] = True if kw.get('width', kw.get('radius')) is not None else False
        self._fixed_size['h'] = True if kw.get('height', kw.get('radius')) is not None else False
        W, H = _info_button(self, **_button_properties._size_opts(self))
        self.cnf['width'] = self.cnf.get('width') if self._fixed_size['w'] else W
        self.cnf['height'] = self.cnf.get('height') if self._fixed_size['h'] else H
        return ('configure', {'width': self.cnf['width'],
                              'height': self.cnf['height']}, None)
    
    def state(self, kw):
        _opt = {}
        if kw.get('state') in 'disabled':
            _opt[1] = ('configure', {'bg': self.cnf.get('disabledbackground'),
                                    'state': 'disabled'}, None)
            _opt[2] = (('itemconfigure', '_txt'), {
                            'disabledfill': kw.get('disabledforeground', 
                            self.cnf.get('disabledforeground')), 
                            'state': 'disabled'}, None)
            _opt[3] = (('itemconfigure', '_activebg'), {'state': 'hidden'}, None)
            _opt[5] = (('itemconfigure', '_img'), {'state': 'disabled'}, None)
            _opt[6] = (('itemconfigure', '_bit'), {'state': 'disabled'}, None)
        elif kw.get('state') == 'normal':
            _bg = self._org_bg
            if self._mouse_state_condition() and self.cnf.get('overbackground'):
                _bg = self.cnf['overbackground']
            _opt[1] = ('configure', {'bg': _bg, 'state': 'normal'}, None)
            _opt[2] = (('itemconfigure', '_txt'), {'state': 'normal', 
                            'fill': self.cnf.get('foreground', 'black')}, None)
            _opt[3] = (('itemconfigure', '_activebg'), {'state': 'hidden'}, None)
            _opt[4] = (('itemconfigure', '_border'), {'state': 'normal'}, None)
            _opt[5] = (('itemconfigure', '_img'), {
                            'image': self.cnf.get('image', '')}, None)
            _opt[6] = (('itemconfigure', '_bit'), {
                            'bitmap': self.cnf.get('bitmap', '')}, None)
        elif kw.get('state') in ('active', 'pressed'):
            _bg = self._org_bg
            if self._mouse_state_condition() and self.cnf.get('overbackground'):
                _bg = self.cnf['overbackground']
            _opt[1] = ('configure', {'bg': _bg, 'state': 'normal'}, None)
            _opt[2] = (('itemconfigure', '_txt'), {'state': 'normal', 
                            'fill': self.cnf.get('activeforeground', 'white')}, None)
            _opt[3] = (('itemconfigure', '_activebg'), {'state': 'normal'}, None)
            _opt[4] = (('itemconfigure', '_border'), {'state': 'hidden'}, None)
            _opt[5] = (('itemconfigure', '_img'), {
                            'image': self.cnf.get('activeimage', '')}, None)
            _opt[6] = (('itemconfigure', '_bit'), {
                            'bitmap': self.cnf.get('activebitmap', '')}, None)
        return _opt

    def takefocus(self, kw):
        _opt = None
        if self['takefocus'] and self['state'] in ('normal', 'active', 'pressed'):
            fn = self._get_functions('takefocus', kw)
            _opt = [self, {'className': 'takefocus', 'sequence':
                            '<FocusIn>', 'func': fn.get('<FocusIn>')},
                            {'className': 'takefocus', 'sequence':
                            '<FocusOut>', 'func': fn.get('<FocusOut>')}]

        elif not self['takefocus'] or self['state'] in 'disabled':
            _opt = [self,
                    {'className': 'takefocus', 'sequence': '<FocusIn>'},
                    {'className': 'takefocus', 'sequence': '<FocusOut>'},
                    (('itemconfigure', '_tf'), {'state': 'hidden'}, None)]
        return _opt

    def text(self, kw):
        if kw.get('textvariable', '') != '':
            kw['text'] = self.cnf['text'] = self.cnf['textvariable'].get()
            cbn = self.cnf['textvariable'].trace_variable('w',
                    self._get_functions('textvariable', kw).get('w'))
            self._var_cb = (self.cnf['textvariable'], cbn)

        elif kw.get('textvariable') == '' and self._var_cb:
            kw['text'] = self.cnf['text'] = self._var_cb[0].get()
            self.cnf.pop('textvariable', None)
            self._var_cb[0].trace_vdelete('w', self._var_cb[1])
            self._var_cb = []

        state = 'disabled' if self.cnf.get('state') in (
            'disabled', 'disable') else 'normal'
        fill = kw.get('fg', self.cnf.get('fg'))
        if self['state'] in ('pressed', 'active'):
            fill = self.cnf.get('activeforeground', 'white')
        _opt = dict(state=state, fill=fill,
                    disabledfill=kw.get(
                        'disabledforeground', self.cnf.get('disabledforeground')),
                    **{i: kw.get(i, self.cnf.get(i)) for i in ('text', 'anchor', 'font', 'justify')})
        return (('itemconfigure', '_txt'), _opt, None)


class _button_items:
    """Internal class.

    Checks and creates (text, image, bitmap, border, 
    bordercolor, takefocus ring*) items."""

    # Utils
    def _active_state(self, val):
        """Internal function."""
        _kw = dict( fill=self.cnf.get('foreground', 'black'),
                    img=self.cnf.get('image', ''),
                    bit=self.cnf.get('bitmap', ''),
                    state='normal')
        if self['state'] in ('pressed', 'active'):
            return dict(fill=self.cnf.get('activeforeground', 'white'),
                        img=self.cnf.get('activeimage', _kw['img']),
                        bit=self.cnf.get('activebitmap', _kw['bit']),
                        state='hidden')[val]
        return _kw[val]
    
    def _bit_img_main(self, opt, *ags, **kw):
        kw['tag'] = '_img' if opt == 'image' else '_bit'
        if (self.cnf.get(opt) != '' or 
                self.cnf.get(str('active'+opt), '') != '' ):
            return self._create(
                opt, (0, 0), {'tag': kw['tag'], 
                opt: _button_items._active_state(self, kw['tag'][1:])})

    # These are main items (function names = item tag name.)
    def _txt(self, *ags, **kw):
        "Text item."
        txt_id = self._create(
            'text', (0, 0), {'text': None, 'tag': '_txt', 
            'fill': _button_items._active_state(self, 'fill')})
        self.cnf['font'] = self.cnf.get('font', self.itemcget(txt_id, 'font'))
        return txt_id

    def _bit(self, *ags, **kw):
        "Bitmap items (bitmap and activebitmap)."
        return _button_items._bit_img_main(self,'bitmap', *ags, **kw)
    
    def _img(self, *ags, **kw):
        "Image item (image and activeimage)."
        return _button_items._bit_img_main(self, 'image', *ags, **kw)
    
    def _bd_color(self, *ags, **kw):
        "Border color item."
        bd_color = self.cnf.get('bordercolor', get_shade(self['bg'], 0.04, 'auto-120'))
        if self._type == 'circle':
            pad = 2
            width = r = int(self.cnf.get('width', 87)/2)  # radius = x = y (in pixels)
            _bd_points = (pad-width, pad-width, r*2+width-pad, r*2+width-pad)
            kw_bd_color = {
                'tag': '_bd_color', 
                'state': _button_items._active_state(self, 'state'),
                'width': width*2,'outline': bd_color}
            return self._create('oval', _bd_points, kw_bd_color)
        _bd_points = (0, -1, self.cnf.get('width', 87), 
                        self.cnf.get('height', 24)+3, 7)  # was 6
        return self.rounded_rect(
            _bd_points, width=6, tag='_bd_color', 
            style='arc', outline=bd_color)
    
    def _border(self, *ags, **kw):
        "Border item."
        bo_color = get_shade(self['bg'], BORDER_INTENSITY, 'auto-120')
        if self._type == 'circle':
            pad = 2
            r = int(self.cnf.get('width', 87)/2)  # radius = x = y
            _bo_points = (pad, pad, r*2-(pad+1), r*2-(pad+1))
            return self._create('oval', _bo_points, {
                'tag': '_border', 'outline': self.cnf.get('bordercolor', bo_color) })
        h = 4
        if int(self['highlightthickness']):
            h += 1
        _bo_points = (2, 2, self.cnf.get('width', 87)-5, self.cnf.get('height', 24)-h, 4)  # was 3
        return self.rounded_rect(
            _bo_points, width=1, outline=bo_color, smooth=1, 
            tag='_border', style='arc', 
            state=_button_items._active_state(self,'state'))
    
    def _tf(self, *ags, **kw):
        "Takefocus highlight ring."
        if self.cnf.get('focusthickness') == 0:
            return
        if self._type == 'circle':
            pad = 1
            width = self.cnf.get('focusthickness', 2)
            r = int(self.cnf.get('width', 87)/2)  # radius = x = y
            _tk_points = (pad+width, pad+width, r*2-width-pad, r*2-width-pad)
            return self._create('oval', _tk_points, {
                    'tag': '_tf', 'width': width,
                    'outline': self.cnf.get('focuscolor', '#81b3f4'), 
                    'state': 'hidden'})

        # takefocuswidth can be changed.
        # Focus line is not on point the line is off when thickness is changed.
        s = w = self.cnf.get('focusthickness', 2)
        diff1 = diff2 = (int(self['highlightthickness'])*2) + (s*2)
        if diff2 == (s*2):
            diff2 -= 1
        
        if s == 1:
            s = 2
            diff2 = (int(self['highlightthickness'])*2) + (s*2)
            diff1 = diff2 + 1
        
        _tk_points = (s+int(self['highlightthickness']),
                        s+int(self['highlightthickness']),
                        self.cnf.get('width',  87)-diff1,
                        self.cnf.get('height', 24)-diff2, 4)
        return self.rounded_rect(
            _tk_points, width=w, style='arc', 
            outline=self.cnf.get('focuscolor', '#81b3f4'), 
            tag='_tf', state='hidden')
    
    # Main function
    def create_items(self, cmd, safe_create=False, **kw):
        _id = []
        for i in BUTTON_ITEMS:
            if self.check_tag(cmd, i, safe_create, kw.get('avoid', [])):
                fn = getattr(_button_items, i)
                _id.append(fn(self, cmd=cmd, safe_create=safe_create, **kw))
        return _id


class _button_functions:
    """Internal class.\n
    Function class for Button widget. Do not call directly."""

    def _mouse_state_condition(self):
        """Internal function.\n
        True if state is normal and cursor is on the widget."""
        con1 = bool(self.cnf.get('state') not in 'disabled')
        con2 = bool(self.winfo_containing(*self.winfo_pointerxy()) == self)
        return con1 and con2
    
    def _make_dictionaries(self, cnf={}, kw={}):
        """Internal function.\n
        Merges kw into cnf and removes None values."""
        kw['bordercolor'] = kw['highlightbackground'] = kw.get(
            'bordercolor', kw.get('highlightbackground'))
        kw = {k: v for k, v in kw.items() if v is not None}
        cnf.update(kw)
        cnf = {k: v for k, v in cnf.items() if v is not None}
        cnf['fg'] = cnf['foreground'] = kw.get(
            'foreground', kw.get('fg', cnf.get('fg', 'black')))
        if cnf.get('textvariable', '') != '':
            cnf['text'] = cnf['textvariable'].get()
        if kw.get('activebackground'):
            cnf.pop('activebackground')
        return cnf, kw

    def _set_trace(self, kw):
        """Internal function."""
        for i in ('activeforeground', 'activebackground', 'bordercolor', 
                  'disabledbackground', 'disabledforeground', 'foreground', 
                  'fg', 'focuscolor', 'highlightbackground', 'overbackground', 
                  'overforeground'):
            if isinstance(kw.get(i), tkinter.Variable):
                var = kw[i]
                cbname = var.trace_variable('w', lambda a, b, c, i=i, var=var,
                                cls=self: cls.config({i: var.get()}))
                if (self, i) in cp._all_traces_colorvar:
                    v, cb = cp._all_traces_colorvar.get((self, i))
                    v.trace_vdelete('w', cb)
                    cp._all_traces_colorvar[(self, i)] = (var, cbname)
                else:
                    cp._all_traces_colorvar[(self, i)] = (var, cbname)
                kw[i] = var.get()
        return kw

    def _get_functions(self, cmds, kw={}):
        """Internal function.\n
        Contains all the required functions."""

        def _borderless_support(cnf={}, **kw):
            _return = self.master._configure('configure', cnf, kw)
            kw = _cnfmerge((cnf, kw))
            if kw.get('bg') or kw.get('background'):
                for i in self._buttons:
                    if i['borderless']:
                        i.cnf['highlightbackground'] = i.cnf['bordercolor'] = i.master['bg']
                        _Canvas._configure(i, ('itemconfigure', '_bd_color'), {
                            'outline': i.master['bg']}, None)
                        _Canvas._configure(i, 'configure', {
                            'highlightbackground': i.master['bg']}, None)
            return _return

        def if_state(fn):
            """Runs function only if state is normal."""
            def wrapper(*a, **k):
                if self.cnf.get('state') not in 'disabled':
                    return fn(*a, **k)
            return wrapper

        def over_img_bit(seq, key):
            """Resets coords for text with overimage/overbitmap."""
            if self.cnf['state'] in ('active', 'pressed'):
                return
            tag = '_img' if key == 'image' else '_bit'
            if seq == 'enter':
                _Canvas._configure(self, ('itemconfigure', tag), 
                                   {key: kw.get('over'+key)}, None)
            elif seq == 'leave':
                _Canvas._configure(self, ('itemconfigure', tag), 
                                   {key: self.cnf.get(key, '')}, None)
            self._set_coords(self._get_options(
                                ('_txt', '_img', '_bit'), self.cnf))
        
        def overbg(seq):
            """Implement overbackground properly."""
            if seq == 'enter':
                _Canvas._configure(self, 'configure', {'bg': kw.get('overbackground')}, None)
                _Canvas._configure(self, ('itemconfigure', '_border'), 
                    {'outline': get_shade(kw.get('overbackground'), 0.1, 'auto-120')}, None)
            elif seq == 'leave':
                self._configure('configure', {'bg': self._org_bg}, None)
        
        def overfg(seq):
            """Handle overforeground, 
            foreground and activeforeground."""
            if self.cnf['state'] in ('active', 'pressed'):
                return
            if seq == 'enter':
                _Canvas._configure(self, ('itemconfigure', '_txt'), 
                    {'fill': self.cnf.get('overforeground', 'black')}, None)
            elif seq == 'leave':
                _Canvas._configure(self, ('itemconfigure', '_txt'), 
                    {'fill': self.cnf.get('fg', 'black')}, None)

        binds = {
            'overrelief': {'<Enter>': if_state(lambda _: self._configure(
                                'configure', {'_relief': kw.get('overrelief')})),
                           '<Leave>': if_state(lambda _: self._configure(
                                'configure', {'_relief': self._rel[0]}))},

            # [issue-3] doesn't change if overrelief is on.
            # [issue-3] (FIXED) using after with 0ms delay fixes the issue. To be safe delay is 1ms.
            'overforeground': {'<Enter>': if_state(lambda _: self.after(1, overfg, 'enter')),
                               '<Leave>': if_state(lambda _: overfg('leave'))},
            
            'overbackground': {'<Enter>': if_state(lambda _: overbg('enter')), 
                               '<Leave>': if_state(lambda _: overbg('leave')) },

            'overimage': {'<Enter>': if_state(lambda _: over_img_bit('enter', 'image')),
                            '<Leave>': if_state(lambda _: over_img_bit('leave', 'image'))},

            'overbitmap': {'<Enter>': if_state(lambda _: over_img_bit('enter', 'bitmap')),
                             '<Leave>': if_state(lambda _: over_img_bit('leave', 'bitmap'))},

            'takefocus': {'<FocusIn>': lambda _: _Canvas._configure(self, (
                            'itemconfigure', '_tf'), {'state': 'normal'}, None),
                          '<FocusOut>': lambda _: _Canvas._configure(self, (
                            'itemconfigure', '_tf'), {'state': 'hidden'}, None)},
        }

        other_functions = {
            'borderless': _borderless_support,

            'textvariable': {'w': lambda *a: _Canvas._configure(self, (
                            'itemconfigure', '_txt'), {'text': self.cnf.get(
                                                       'textvariable').get()}, None)},

            '_activebg': lambda: _on_press_color(*self._get_options(
                                                 'activebackground', self.cnf)),
        }
        funcs = _cnfmerge((binds, other_functions))
        if isinstance(cmds, (list, tuple)):
            value = funcs.copy()
            for i in cmds:
                value = value.get(i, {})
            return value
        return funcs.get(cmds, {})

    def _get_options(self, cmd, cnf={}, **kw):
        """Internal function."""
        kw = _cnfmerge((cnf, kw))
        # If more than one commands are given.
        # returns Union[list, dict, tuple(list, dict)]
        if isinstance(cmd, (list, tuple)):
            opts = None
            ags, cnf = [], {}
            for i in cmd:
                opts = self._get_options(i, kw)
                if isinstance(opts, (list, tuple)):
                    ags.append(opts)
                elif isinstance(opts, dict):
                    cnf.update(opts)
            if isinstance(opts, (list, tuple)) and isinstance(opts, dict):
                return ags, cnf
            return ags or cnf

        if cmd in _button_properties.__dict__:
            return getattr(_button_properties, cmd)(self, kw)

    def _set_coords(self, cnf={}, **kw):
        """Internal function.\n
        Set Coordinates of the items."""
        kw = _cnfmerge((cnf, kw))
        r = [self.coords(tag, *kw[tag]) for tag in
                ('_txt', '_img', '_bit', '_bd_color',
                 '_border', '_tf') if kw.get(tag)]
        for i in ('_txt', '_bit', '_img'):
            if self.coords(i):
                self._set_anchor(self.cnf['anchor'], i)
        return r

    def _set_configure(self, options):
        """Internal function.\n
        Configures and binds according to the given options."""
        if options and isinstance(options, dict):
            return [self._set_configure(i) for i in options.values()]
        if not (options and isinstance(options[1], dict)):
            return

        con1 = bool(isinstance(options[0], tuple))
        if options[0] == 'configure' or (con1 and len(options[0]) > 1):
            # itemconfigure and configure
            for i in ('_txt', '_bit', '_img'):
                if self.coords(i) and options[1].get('anchor'):
                    self._set_anchor(options[1].pop('anchor', 'center'), i)
            _Canvas._configure(self, *options)

        if isinstance(options[0], tkinter.Misc):
            if isinstance(options[-1], (tuple, list)):
                # binds, itemconfigure and configure
                binds, conf = options[:-2], options[-1]
                _Canvas._configure(self, *conf)
                return _bind(*binds)
            if options[1].get('tag') == '_activebg':
                return _on_press_color(*options)    # _on_press_color
            return _bind(*options)  # binds

    def _configure1(self, cnf={}, **kw):
        """Internal Function.
        This function configure all the resources of 
        the Widget and save it to the class dict."""
        self.cnf, kw = self._make_dictionaries(
            self.cnf, self._set_trace(
                check_param(self, 'button', _cnfmerge((cnf, kw)))
            )
        )
        # Checks the items
        _button_items.create_items(self, 'check')
        # >.<
        for opt in ('overbackground', 'activebitmap', 'activeimage',
                    'bitmap', 'fg', 'bordercolor', 'borderless', 'image',
                    'overrelief', 'foreground', 'state', 'overforeground',
                    'focuscolor', 'focusthickness', 'activebackground', 
                    'activeforeground', 'compound'):
            if kw.get(opt) is not None:
                self._set_configure(self._get_options(opt, kw))
                if ((opt == 'state' and kw.get('compound', self.cnf.get('compound')))
                    or opt == 'compound'):
                    self._set_coords(**self._get_options(('_txt', '_img', '_bit'), kw))
        # Size
        if bool({'text', 'font', 'textvariable', 'image', 'bitmap', 'compound',
                 'padx', 'pady', 'width', 'height', 'activeimage', 'radius',
                 'activebitmap',}.intersection(set(kw))):
            self._set_configure(self._get_options('size', kw))
        # Text
        if {'text', 'anchor', 'font', 'justify', 'textvariable'}.intersection(set(kw)):
            self._set_configure(self._get_options('text', kw))
        # Takefocus
        cond1 = bool(self.itemcget('_tf', 'state') != 'hidden' and
                     self['state'] not in 'disabled')
        cond2 = bool(self.itemcget('_tf', 'state') != self['state'])
        if cond1 or cond2:
            self._set_configure(self._get_options('takefocus', kw))
        # Line border: This will darken the border around the button.
        if get_shade(self['bg'], BORDER_INTENSITY, 'auto-120') != self.itemcget('_border', 'outline'):
            self._set_configure(self._get_options('_border', kw))

    def _focus_in_out(self, intensity):
        """Internal function.\n
        Focus in and focus out effect maker."""
        self._main_win = self.winfo_toplevel()
        if not self._main_win.winfo_exists():
            return

        def _chngIn(evt):
            """Internal function."""
            try:
                if self.focus_get() is None:
                    self._tmp_bg = self['bg']
                    # [BUG] winfo_rgb doesn't read mac system colors.
                    color1 = self.winfo_rgb(self['highlightbackground'])
                    color1 = [int((int(i/257) + 255)/2) for i in color1]
                    color1 = '#%02x%02x%02x' % (color1[0], color1[1], color1[2])
                    color2 = get_shade(color1, intensity, 'auto-120')
                    _Canvas._configure(self, 'configure', {'bg': color1}, None)
                    _Canvas._configure(self, ('itemconfigure', '_border'),
                            {'outline': color2}, None)
                if self.focus_get():
                    if getattr(self, '_tmp_bg', False):
                        _Canvas._configure(self, 'configure',{'bg': self._tmp_bg}, None)
                    color = get_shade(self['bg'], intensity, 'auto-120')
                    _Canvas._configure(self, ('itemconfigure', '_border'),
                             {'outline': color}, None)
            except KeyError: 
                pass

        _bind(self._main_win, 
              {'className': 'focus%s' % str(self),
               'sequence': '<FocusIn>', 'func': _chngIn},
              {'className': 'focus%s' % str(self),
               'sequence': '<FocusOut>', 'func': _chngIn})
        return self._main_win

    def _relief(self, cnf, kw={}):
        """Internal function.\n
        Make overrelief and Relief work together."""
        kw = _cnfmerge((cnf, kw))
        cond1 = kw.get('overrelief', self.cnf.get('overrelief', '')) != ''
        if kw.get('relief') is not None and cond1:
            self._rel = (kw['relief'], True)
        if kw.get('_relief') is not None and kw.get('relief') is not None:
            self._rel = (kw['relief'], True)
            kw.pop('_relief', None)
        elif kw.get('_relief'):
            kw['relief'] = kw.pop('_relief')
        return kw

    def _set_size(self, evt=None):
        """Internal function.\n
        This will resize everything that is in the button"""
        if evt.width == self._size[0] and evt.height == self._size[1]:
            return
        # [issue-9] (NOT FIXED) On resizing the window the circlebutton doesn't resize properly, 
        #           current fix doesn't work properly.
        if self._type == 'circle' and evt.width != evt.height:
            if evt.width > evt.height:
                evt.width = evt.height
                # tkinter.Canvas.config(self, width=evt.width)
            else:
                evt.height = evt.width
                # tkinter.Canvas.config(self, height=evt.width)
            self.cnf['radius'] = evt.width
        self._size = (self.cnf['width'], self.cnf['height']) = (evt.width, evt.height)
        self.delete('_activebg')
        for i in self._after_IDs:
            self.after_cancel(self._after_IDs[i])
        _button_items.create_items(
            self, 'create', avoid=('_txt', '_img', '_bit'), safe_create=True)
        self._set_coords(self._get_options(('_txt', '_img', '_bit'), self.cnf))
        self._after_IDs[1] = self.after(1, self._get_functions('_activebg'))
        for t in ('_img', '_bit', '_txt', '_bd_color', '_border', '_tf'):
            self.tag_raise(t)
        self._after_IDs[2] = self.after(1, self._configure1)
        cur_focus = self.master.focus_get()
        if cur_focus and self.master != cur_focus and cur_focus != self:
            cur_focus.focus()
        else:
            self.master.focus()
            # [issue-7] Need a better fix to get focus back to the button if it has it previously
            if cur_focus == self:
                # [issue-7] (fixed) using after with 1ms delay solves the issue.
                self._after_IDs[3] = self.after(1, self.focus)

    def _active(self, value):
        """Internal function.\n 
        Do not call directly. Changes appearance when active."""

        def check_active_img_bit(op=''):
            tag = '_bit' if 'bit' in op else '_img'
            if self['activeimage'] != '':
                _Canvas._configure(self, ('itemconfigure', tag), 
                    {'image': self[op+'image']}, None)
            elif self['activebitmap'] != '':
                 _Canvas._configure(self, ('itemconfigure', tag), 
                    {'bitmap': self[op+'bitmap']}, None)

        if value in ('on_press', 'on_enter') or value is True:
            if self['state'] != 'pressed' :
                if value == 'on_press': 
                    self.cnf['_state'] = self.cnf['state']
                self.cnf['state'] = 'active'
                _Canvas._configure(self, ('itemconfigure', '_activebg'), 
                                   {'state': 'normal'}, None)
            _Canvas._configure(self, ('itemconfigure', '_border'), 
                               {'state': 'hidden'}, None)
            _Canvas._configure(self, ('itemconfigure', '_txt'), 
                               {'fill': self['activeforeground']}, None)
            check_active_img_bit('active')

        elif value in ('on_leave', 'on_release') or value is False:  # When not active (False)
            if self['state'] != 'pressed':
                if self.cnf.get('_state'): 
                    self.cnf['state'] = self.cnf['_state']
                if value == 'on_release':
                    self.cnf.pop('_state', None)
                    if self['state'] == 'active': 
                        self['state'] = 'normal'
                _Canvas._configure(self, ('itemconfigure', '_activebg'), 
                                   {'state': 'hidden'}, None)
                _Canvas._configure(self, ('itemconfigure', '_border'), 
                                   {'state': 'normal'}, None)
                fill = self['fg']
                if self._mouse_state_condition() and self['overforeground'] != '':
                    fill = self['overforeground']
                _Canvas._configure(self, ('itemconfigure', '_txt'), 
                                   {'fill': fill}, None)
                check_active_img_bit()

        self._set_coords(self._get_options(('_txt', '_img', '_bit'), self.cnf))

    def _on_press(self, *ags):
        """Internal function. When button is pressed <Button-1>"""
        self._rpin = None
        self._rpinloop = True

        def cmd(*a):
            """trigger function callback."""
            _bind(self, className='button_command', sequence='<ButtonRelease-1>')
            if self.cnf.get('repeatdelay', 0) and self.cnf.get('repeatinterval', 0) and self._rpinloop:
                self._rpin = self.after(self.cnf.get('repeatinterval', 0), cmd)
            if self.cnf.get('command'):
                self.cnf['command']() 

        def on_enter(*a):
            """Internal function.\n
            Enables when pressed and cursor
            is moved back on button."""
            self._active('on_enter')
            if self.cnf.get('repeatdelay', 0) and self.cnf.get('repeatinterval', 0):
                self._rpinloop = True
                cmd()
            _bind(self, className='button_command', sequence='<ButtonRelease-1>',
                  func=lambda *a: self.after(0, cmd), add='+')

        def on_leave(*a):
            """Internal function.\n
            Disables/Cancels when pressed and cursor
            is moved away from the button."""
            self._active('on_leave')                
            if self.cnf.get('repeatdelay', 0) and self.cnf.get('repeatinterval', 1):
                self._rpinloop = False
                self.after_cancel(self._rpin)
            _bind(self, className='button_command', sequence='<ButtonRelease-1>')

        if self['state'] not in 'disabled':
            self.focus_set()
            self._active('on_press')            
            if self.cnf.get('repeatdelay', 0) and self.cnf.get('repeatinterval', 1):
                self._rpin = self.after(self.cnf.get('repeatdelay', 0), cmd)
            _bind(self, {'className': 'on_press_enter', 'sequence': '<Enter>', 'func': on_enter},
                  {'className': 'on_press_leave', 'sequence': '<Leave>', 'func': on_leave},
                  {'className': 'button_command', 'sequence': '<ButtonRelease-1>', 'func': cmd})

    def _on_release(self, *ags):
        """Internal function. When button is released <ButtonRelease-1>"""
        if self['state'] in 'disabled': 
            return
        self._active('on_release')
        self._rpinloop = False
        if getattr(self, '_rpin', None): 
            self.after_cancel(self._rpin)
        _bind(self, {'className': 'on_press_enter', 'sequence': '<Enter>'},
              {'className': 'on_press_leave', 'sequence': '<Leave>'},
              {'className': 'button_command', 'sequence': '<ButtonRelease-1>'})

    def _compound(self, flag, width, height):
        """Internal function.\n
        Use `compound = 'left'/'right'/'top'/'bottom'` to configure."""
        _PiTag = ''
        if self.cnf.get('image', self.cnf.get('activeimage')):
            _PiTag = '_img'
        elif self.cnf.get('bitmap', self.cnf.get('activebitmap')):
            _PiTag = '_bit'
        _im_size = self.bbox(_PiTag)
        _txt_size = self.bbox('_txt')
        if _im_size and _txt_size:
            W_im = _im_size[2] - _im_size[0]
            H_im = _im_size[3] - _im_size[1]
            W_txt = _txt_size[2] - _txt_size[0]
            H_txt = _txt_size[3] - _txt_size[1]
            if flag == 'bottom':
                width = (width/2, width/2)
                height = (height/2-H_im/2, height/2+H_txt/2)
            elif flag == 'top':
                width = (width/2, width/2)
                height = (height/2+H_im/2, height/2-H_txt/2)
            elif flag == 'right':
                width = (width/2-W_im/2, width/2+W_txt/2)
                height = (height/2, height/2)
            elif flag == 'left':
                width = (width/2+W_im/2, width/2-W_txt/2)
                height = (height/2, height/2)        
            if isinstance(height, tuple):
                return {'_txt': (width[0], height[0]),
                            _PiTag: (width[1], height[1])}

    def _set_anchor(self, anchor, item):
        """Internal function.\n
        Sets the anchor position from (n, ne, e, se, s, sw, w, nw, or center)."""
        if self.cnf.get('compound') is not None:
            return
        bbox = self.bbox(item) or [0, 0, 0, 0]
        item_width = bbox[2] - bbox[0]
        item_height = bbox[3] - bbox[1]
        default_padx = 2 + int(item_width/2)
        default_pady = 0 + int(item_height/2)
        width = self.cnf['width'] = self.winfo_width()
        height = self.cnf['height'] = self.winfo_height()

        # Center
        x = width / 2
        y = height / 2

        if anchor in ('n', 'nw', 'ne'):
            x = int(width/2)
            y = default_pady + 0
            if anchor == 'nw':
                x = default_padx + 0
            elif anchor == 'ne':
                x = default_padx + (width - item_width) - 4
        
        elif anchor in ('s', 'sw', 'se'):
            x = int(width/2)
            y = default_pady + (height - item_height)
            if anchor == 'sw':
                x = default_padx + 0
            elif anchor == 'se':
                x = default_padx + (width - item_width) - 4
        
        elif anchor == 'e':
            x = default_padx + (width - item_width) - 4
                
        elif anchor == 'w':
            x = default_padx + 0
        return self.coords(item, x, y)


class ButtonBase(_Canvas, _button_functions):
    """Internal class used for tkinter macos Buttton"""

    _buttons = []  # list of all buttons

    def __init__(self, _type=None, master=None, cnf={}, **kw):
        kw = self._set_trace(_cnfmerge((cnf, kw)))
        kw = {k: v for k, v in kw.items() if v is not None}
        self._type = _type  # button type (circle, normal)
        self._after_IDs = {} # _after_IDs
        self._fixed_size = {'w': False, 'h': False}
        self._var_cb = None
        self.cnf = {}
        
        cnf = {}
        for i in kw.copy().keys():
            if i in BUTTON_FEATURES:
                cnf[i] = kw.pop(i, None)

        cnf['fg'] = cnf['foreground'] = cnf.get('fg', cnf.get('foreground', SYSTEM_DEFAULT_FG))
        cnf['anchor'] = cnf.get('anchor', 'center')
        cnf['justify'] = cnf.get('justify', 'center')
        cnf['borderless'] = cnf.get('borderless', False)
        cnf['disabledforeground'] = cnf.get('disabledforeground', 'grey')
        cnf['disabledbackground'] = cnf.get('disabledbackground', 'grey')
        cnf['state'] = cnf.get('state', 'normal')
        cnf['activeforeground'] = cnf.get('activeforeground', 'white')
        cnf['activebackground'] = cnf.get('activebackground', ("#4b91fe", "#055be5"))
        cnf['focuscolor'] = cnf.get('focuscolor', '#81b3f4')
        cnf['focusthickness'] = cnf.get('focusthickness', 2)
        cnf['compound'] = cnf.get('compound', None)
        cnf['padx'] = cnf.get('padx', 1)
        cnf['pady'] = cnf.get('pady', 1)
        cnf['repeatdelay'] = cnf.get('repeatdelay', 0)
        cnf['repeatinterval'] = cnf.get('repeatinterval', 1)
        cnf['wraplength'] = cnf.get('wraplength', 0)
        cnf['underline'] = cnf.get('underline', -1)

        for i in ('activebitmap', 'activeimage', 'bitmap', 'command', 
                  'image', 'overbackground', 'overforeground', 'overrelief', 
                  'text', 'textvariable'):
            cnf[i] = cnf.get(i, '')

        if self._type == 'circle':
            cnf['radius'] = int(kw.pop('radius', 35)) 
            ra = int(cnf['radius']*2 + 4)
            kw['width'] = kw['height'] = kw.get('width', kw.get('height', ra))
        else:
            kw['width'] = kw.get('width', 87)
            kw['height'] = kw.get('height', 24)

        kw['takefocus'] = kw.get('takefocus', 1)
        kw['bg'] = kw.pop('bg', kw.pop('background', SYSTEM_DEFAULT_BG))
        kw['highlightthickness'] = kw.get('highlightthickness', 0)

        _Canvas.__init__(self, master=master, **kw)
        self.cnf = check_param(self, 'button', **cnf)
        self.cnf['bordercolor'] = self.cnf['highlightbackground'] = self.cnf.get(
            'bordercolor', self.cnf.get('highlightbackground', get_shade(self['bg'], 0.04, 'auto-120')))

        self._buttons.append(self)
        self._size = (self.winfo_width(), self.winfo_height())
        _button_items.create_items(self, 'create', safe_create=True)
        self._org_bg = self['bg']
        if kw.get('relief') is not None:
            self._rel = (kw['relief'], True)
        else:
            self._rel = ('flat', False)

        _bind(self,
              {'className': 'button_release',
                  'sequence': '<ButtonRelease-1>', 'func': self._on_release},
              {'className': 'button_press',
                  'sequence': '<Button-1>', 'func': self._on_press},
              {'className': 'set_size', 'sequence': '<Configure>', 'func': self._set_size})

        self._focus_in_out(BORDER_INTENSITY)
        self._configure1(self.cnf)

    def _configure(self, cmd, cnf=None, kw=None):
        'Internal function to configure the inherited class'
        kw = self._relief(cnf, kw)
        cnf = {}
        for i in list(kw):
            if ((i in BUTTON_FEATURES)
                or (i == 'radius' and self._type == 'circle')):
                cnf[i] = kw.pop(i, None)
        _return = _Canvas._configure(self, cmd, None, kw)
        if kw.get('bg') or kw.get('background'):
            self._org_bg = self['bg']
        self._configure1(cnf)
        if _return is not None and isinstance(_return, dict):
            _return.update(self.cnf)
            _return['compound'] = self.cnf.get('compound', 'none')
            for k in list(_return):
                if k not in self.keys():
                    _return.pop(k)
        return _return

    def cget(self, key):
        """Return the resource value for a KEY given as string."""
        need_string = ('textvariable', 'image', 'activeimage')
        need_int = ('borderwidth', 'bd', 'highlightthickness')
        if key == 'radius' and self._type == 'circle':
            return self.cnf.get('radius')
        if key in BUTTON_FEATURES:
            if key in need_string and self.cnf.get(key):
                return str(self.cnf.get(key))
            if key == 'command' and self.cnf.get(key) is None:
                return 'none'
            return self.cnf.get(key)
        value = _Canvas.cget(self, key)
        if key in need_int:
            return int(value)
        return value
    __getitem__ = cget

    def keys(self):
        """Return a list of all resource names of this widget."""
        return sorted(
            set([*BUTTON_PROPERTIES, *BUTTON_FEATURES])
        )
    
    def invoke(self):
        """Invoke the command associated with the button.

        The return value is the return value from the command,
        or an empty string if there is no command associated with
        the button. This command is ignored if the button's state
        is disabled.
        """
        if (self['state'] not in ('disable', 'disabled') 
                and self.cnf.get('command')):
            return self.cnf['command']()

    @cp._colorvar_patch_destroy
    def destroy(self):
        """Destroy this and all descendants widgets. This will
        end the application of this Tcl interpreter."""
        if self._main_win.winfo_exists():
            _bind(self._main_win,
                {'className': 'focus%s' % str(self), 'sequence': '<FocusIn>'},
                {'className': 'focus%s' % str(self), 'sequence': '<FocusOut>'})
        for i in self._after_IDs:
            self.after_cancel(self._after_IDs[i])
        if self.cnf.get('textvariable', '') != '':
            self.configure(textvariable='')
        if self in self._buttons:
            self._buttons.remove(self)
        return _Canvas.destroy(self)


#                       Copyright 2021 Saad Mairaj
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tkmacosx.basewidgets.marquee_base import MarqueeBase


class Marquee(MarqueeBase):
    """Marquee widget."""

    def __init__(self, master=None, cnf={}, **kw):
        """Use `Marquee` for creating scrolling text which moves from 
        right to left only if the text does not fit completely.

        ### Args:
        - `text`: Give a string to display.
        - `font`: Font of the text.
        - `fg`: Set foreground color of the text.
        - `fps`: Set fps(frames per seconds).
        - `left_margin`: Set left margin to make text move further in right direction before reset.
        - `initial_delay`: Delay before text start to move.
        - `end_delay`: Delay before text reset.
        - `smoothness`: Set the smoothness of the animation.

        ### Example: 
        ```
        root=tk.Tk()
        marquee = Marquee(root, 
                        text='This text will move from right to left if does not fit the window.')
        marquee.pack()
        root.mainloop()
        ```

        ### To move the text when cursor is over the text then follow the below example.
        ```
        text = "Please hover over the text to move it. "
               "This text will move only if the cursor "
               "hovers over the text widget."
        root = tk.Tk()
        m = tkm.Marquee(root, bg='lightgreen', text=text)
        m.pack()
        m.stop(True)
        m.bind('<Enter>', lambda _: m.play())
        m.bind('<Leave>', lambda _: m.stop())
        root.mainloop()
        ```"""
        MarqueeBase.__init__(self, master=master, cnf=cnf, **kw)

    def reset(self):
        """Reset the text postion."""
        self._reset(True)
        self._stop_state = False

    def stop(self, reset=False):
        """Stop the text movement."""
        if reset:
            self.reset()
        self._stop_state = True
        self.after_cancel(self.after_id)
        self.after_id = ' '

    def play(self, reset=False):
        """Play/continue the text movement."""
        if not self._stop_state and not reset:
            return
        self._stop_state = False
        if reset:
            self.reset()
        self._animate()



#                       Copyright 2021 Saad Mairaj
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from tkmacosx.basewidgets.colorscale_base import (ColorscaleBase, HEX, RGB)


class Colorscale(ColorscaleBase):
    "Colorscale widget."

    def __init__(self, master=None, cnf={}, **kw):
        """
        ## Color Scale.
        This is ColorScale alternate to colorchooser of tkinter. 

        ### Args: 
        - `value`: Get either 'RGB' or 'HEX'.
        - `command`: callback function with an argument.
        - `orient`: Set the orientation.
        - `mousewheel`: Set mousewheel to scroll marker.
        - `variable`: Give tkinter variable (`StringVar`).
        - `showinfo`: Shows hex or rbg while selecting color.
        - `gradient`: Take tuple of two colors or default.
        - `showinfodelay`: Delay before the showinfo disappears (in ms).
        """
        ColorscaleBase.__init__(self, master=master, cnf=cnf, **kw)



#                       Copyright 2021 Saad Mairaj
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import pickle
import tkinter


def SaveVar(var, master=None, value=None, name=None, filename='data.pkl'):
    """Save tkinter variable data in a pickle file and load the 
    same value when the program is executed next time. 

    #### If the content of the file changes, it might not load correct values \
        to the assigned variables. To avoid this issue use `name` to \
        refer to the exact assigned values later.

    ### Args:
    - `var`: Give the `tkinter.Variable` class like (`tk.StringVar`, `tk.IntVar`).
    - `master`: Parent widget.
    - `value`: Set value.
    - `name`: Set a name to group variables or to refer to assigned value when loaded.
    - `filename`: Set the name of the save file. (To make the file invisible in the \
            directory start the name of the file with "." like ".cache-savevar")

    ### Return:
    - returns the tk.Variable instance passed to `var` argument.

    ### Example:
        root = tk.Tk()
        var1 = SaveVar(tk.StringVar,'Enter Username','Var1','.cache-savevar')
        var2 = SaveVar(tk.StringVar,'Enter Password','Var2','.cache-savevar')
        tk.Entry(root,textvariable=var1).pack()
        tk.Entry(root,textvariable=var2).pack()
        root.mainloop()"""

    def update_val(*args):
        """Internal function for updating the value for variable"""
        try:    # try/except , if the file doesn't exists.
            open1 = open(filename, 'rb')
            tmpdict = pickle.load(open1)  # load saved dictionary data.
            # Block of code to check for the right value.
            if tmpdict.get(str(var)):
                old, default = tmpdict.get(str(var))
                new = var.get()
                if new != default:
                    var.set(new)
                elif not startup[0]:
                    var.set(default)
                else:
                    var.set(old)
            tmpdict.update({str(var): (var.get(), defaultval)})
            open1.close()
        except FileNotFoundError:
            tmpdict = {}
            tmpdict[str(var)] = (var.get(), defaultval)

        open2 = open(filename, 'wb')
        pickle.dump(tmpdict, open2)
        startup[0] = False
        open2.close()

    startup = [True]
    if not(filename.endswith('.pickle') or filename.endswith('.pkl')) \
            and not filename.startswith('.'):
        filename = filename+'.pkl'
    var = var(master=master, value=value, name=name)
    defaultval = var.get()  # get a default value of the variable
    update_val()
    for mode, cbname in (var.trace_vinfo()):
        if mode[0] == 'w' and update_val.__name__ in cbname:
            try:
                var.trace_vdelete('w', cbname)
            except tkinter.TclError:
                pass
    var.trace_variable('w',  update_val)
    return var



#                       Copyright 2021 Saad Mairaj
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import re
import sys
import colour
import tkinter
from tkinter import ttk
from tkmacosx.utils.check_parameter import Check_Comman_Parameters, pixels_conv


SYSTEM_DEFAULT_BG = "white"
SYSTEM_DEFAULT_FG = "black"
if sys.platform == 'darwin':
    SYSTEM_DEFAULT_BG = "systemWindowBackgroundColor"
    SYSTEM_DEFAULT_FG = "systemTextColor"


def _agsmerge(args):
    """Internal functions.\n
    Merges lists/tuples."""
    if not isinstance(args, (tuple, list)):
        return args
    a = []
    for i in args:
        if isinstance(i, (tuple, list)):
            a.extend(i)
        else:
            a.append(i)
    return a


def _bind(cls=None, *ags, **kw):
    """Internal function.\n
    Binds and unbinds sequences with any name given as className."""
    cls = cls or kw.pop('cls', ags.pop(0))
    if ags:
        return [_bind(cls=cls, **i) for i in ags]
    classname = kw['className'] + str(cls)
    bindtags = list(cls.bindtags())
    if classname in bindtags:
        bindtags.remove(classname)
    if kw.get('func'):
        _bind(cls, className=kw['className'], sequence=kw['sequence'])
        bindtags.append(classname)
        cls.bindtags(tuple(bindtags))
        return cls.bind_class(classname, sequence=kw['sequence'],
                              func=kw['func'], add=kw.get('add', '+'))
    cls.bindtags(tuple(bindtags))
    cls.unbind_class(classname, kw['sequence'])


def _cnfmerge(cnfs):
    """Internal function."""
    if isinstance(cnfs, (type(None), str, dict)):
        return cnfs
    cnf = {}
    for c in tkinter._flatten(cnfs):
        try:
            if isinstance(c, dict):
                cnf.update(c)
        except (AttributeError, TypeError) as msg:
            print("_cnfmerge: fallback due to:", msg)
            for k, v in c.items():
                cnf[k] = v
    return cnf


def _info_button(cls, cnf={}, **kw):
    """Internal Function.\n
    This function takes essentials parameters to give
    the approximate width and height accordingly. \n
    It creates a ttk button and use all the resources given
    and returns width and height of the ttk button, after taking
    width and height the button gets destroyed also the custom style."""
    kw = _cnfmerge((cnf, kw))
    if kw.get('bitmap'):
        tmp = tkinter.Button(cls, **kw)
    else:
        kw['style'] = '%s.TButton' % cls
        _style_tmp = ttk.Style(master=cls)
        _style_tmp.configure(kw['style'], font=kw.pop('font', None))
        _style_tmp.configure(kw['style'], padding=(
            kw.pop('padx', 0), kw.pop('pady', 0)))
        tmp = ttk.Button(cls, **kw)
        # [issue-2] Need fix --- doesn't really delete the custom style
        del _style_tmp
    geo = [tmp.winfo_reqwidth(), max(24, tmp.winfo_reqheight()-4)]
    tmp.destroy()
    return geo


def _on_press_color(cls=None, cnf={}, **kw):
    """Internal function. Do not call directly.\n
    Give gradient color effect used for activebackground.
    Returns ids"""
    kw = _cnfmerge((cnf, kw))
    cls = kw.get('cls', cls)
    w = cls.cnf.get('height', cls.winfo_width())
    h = cls.cnf.get('height', cls.winfo_height())
    tag = kw.get('tag', 'press')
    state = kw.get('state', 'normal' if cls.cnf.get('state')
                   in ('active', 'pressed') else 'hidden')
    if not cls:
        raise ValueError('Could not refer to any class instance "cls".')
    if kw.get('color') is None:
        kw.pop('color', None)
    width = cls.coords(tag) or 0
    if isinstance(width, (list, tuple)) and len(width) > 3:
        width = int(width[2])
    all_activebg_ids = cls.find('withtag', tag)
    cond1 = bool(h == len(all_activebg_ids))
    cond2 = bool(w == width)
    # [issue-1] Need a better approach for getting "cond3"
    cond3 = bool(kw.get('color', (False,)) == cls.cnf.get('activebackground'))
    if cond1 and cond2 and cond3 and not kw.get('force_create', False):
        return
    # This is the default accent color for mac.
    cr = cls.cnf['activebackground'] = kw.get(
        'color', cls.cnf.get('activebackground', ("#4b91fe", "#055be5")))
    cls.delete(tag)
    ids = []
    height = kw.get('height', h)
    if isinstance(cr, (tuple, list)) and None in cr:
        cr = list(cr)
        cr.remove(None)
        cr = cr[0]
    if not isinstance(cr, (tuple, list)):
        cr_list = (cr,) * height
    else:
        cr = (colour.Color(cr[0]), colour.Color(cr[1]))
        cr_list = tuple(
            cr[0].range_to(cr[1], height)
        )
    for i in range(height):
        ags = (0, i, kw.get('width', w), i)
        cnf = {'fill': cr_list[i], 'tag': tag, 'state': state}
        ids.append(cls._create('line', ags, cnf))
    cls.tag_lower(tag)     # keep the tag lowest
    return tuple(ids)


class _Canvas(tkinter.Widget):
    """Internal Class."""

    def __init__(self, master=None, cnf={}, **kw):
        tkinter.Widget.__init__(self, master, 'canvas', cnf, kw)

    def find(self, *args):
        """Internal function."""
        return self._getints(
            self.tk.call((self._w, 'find') + args)) or ()

    def bbox(self, *args):
        """Return a tuple of X1,Y1,X2,Y2 coordinates for a rectangle
        which encloses all items with tags specified as arguments."""
        return self._getints(
            self.tk.call((self._w, 'bbox') + args)) or None

    def coords(self, *args):
        """Return a list of coordinates for the item given in ARGS."""
        return [self.tk.getdouble(x) for x in
                self.tk.splitlist(
            self.tk.call((self._w, 'coords') + args))]

    def _create(self, itemType, args, kw):  # Args: (val, val, ..., cnf={})
        """Internal function."""
        args = tkinter._flatten(args)
        cnf = args[-1]
        if isinstance(cnf, (dict, tuple)):
            args = args[:-1]
        else:
            cnf = {}
        return self.tk.getint(self.tk.call(
            self._w, 'create', itemType,
            *(args + self._options(cnf, kw))))

    def delete(self, *args):
        """Delete items identified by all tag or ids contained in ARGS."""
        self.tk.call((self._w, 'delete') + args)

    def itemcget(self, tagOrId, option):
        """Return the resource value for an OPTION for item TAGORID."""
        return self.tk.call(
            (self._w, 'itemcget') + (tagOrId, '-'+option))

    def check_tag(self, cmd, tag, safe_create=False, avoid=[]):
        """Internal function.\n
        If `cmd="check"` and the tag does not exist then
        the tag is created, but if `cmd="create"` and
        safe_create=True this will delete the tag if exists
        and creates a new tag  with same settings."""
        c = True
        if cmd == 'check':
            c = False
        if cmd not in ('create', 'check'):
            raise ValueError(
                '%s is not a valid command! Takes -create, -check' % cmd)
        cond1 = bool(not self.find('withtag', tag) or c)
        cond2 = bool(tag not in avoid)
        if safe_create and cond1 and cond2:
            self.delete(tag)
        return cond1 and cond2

    def tag_lower(self, *args):
        """Lower an item TAGORID given in ARGS
        (optional below another item)."""
        self.tk.call((self._w, 'lower') + args)

    def tag_raise(self, *args):
        """Raise an item TAGORID given in ARGS
        (optional above another item)."""
        self.tk.call((self._w, 'raise') + args)

    def rounded_rect(self, ags=(), *args, **kw):
        """Internal function."""
        x, y, w, h, c = _agsmerge((ags, args))
        ids = []
        cnf = dict(kw)
        for i in ('extent', 'start', 'style'):
            cnf.pop(i, None)
        for i in ('joinstyle', 'smooth', 'slinesteps'):
            kw.pop(i, None)
        points = (  # Arc points:-
            (x, y, x+2*c, y+2*c),
            (x, y+h-2*c, x+2*c, y+h),
            (x+w-2*c, y+h-2*c, x+w, y+h),
            (x+w-2*c, y, x+w, y+2*c),
            # Polygon points:-
            (x+c, y, x+w-c, y),
            (x+c, y+h, x+w-c, y+h),
            (x, y+c, x, y+h-c),
            (x+w, y+c, x+w, y+h-c))

        for i, point in enumerate(points):
            if i <= 3:
                kw['start'] = 90*(i+1)
                ids.append(self._create('arc', point, kw))
            else:
                ids.append(self._create('polygon', point, cnf))
        return tuple(ids)

    def _rounded(self, ags=(), *args, **kw):
        """Internal function."""
        x1, y1, x2, y2, c = _agsmerge((ags, args))
        ids = []
        points = (  # Arc points:-
            (x2-c-1, y1, x2-1, y1+c),
            (x1, y1, x1+c, y1+c),
            (x1, y2-c-1, x1+c, y2-1),
            (x2-c, y2-c, x2-1, y2-1),
            # Rectangle points:-
            (x1+c/2, y1, x2-c/2, y2),
            (x1, y1+c/2, x2, y2-c/2))

        kw['start'], kw['outline'] = 0, ''
        for i, point in enumerate(points):
            if i <= 3:
                ids.append(self._create('arc', point, kw))
                kw['start'] += 90
            else:
                kw.pop('start', None)
                kw['width'] = 0
                ids.append(self._create('rectangle', point, kw))
        return tuple(ids)


def check_appearance(cmd='defaults read -g AppleInterfaceStyle'):
    """### Checks DARK/LIGHT mode of macos. Returns Boolean.
    #### Args:
    - `cmd`: Give commands. Like to check DARK/LIGHT mode \
            the command is `'defaults read -g AppleInterfaceStyle'` .
    """
    import subprocess
    cmd = 'defaults read -g AppleInterfaceStyle'
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE, shell=True)
    return bool(p.communicate()[0])


def check_function_equality(func1, func2):
    """Checks if two functions are same."""
    return func1.__code__.co_code == func2.__code__.co_code


def check_light_dark(value, intensity=110):
    """Tells if the given RGB or HEX code is lighter or darker color.

    Args:
        value [str, tuple]: Give sequence of RBG values and hexcode.
        intensity (int): The measurable amount of a brightness.

    Return:
        str('white') / str('black')"""
    if isinstance(value, str) and value.startswith('#'):
        value = hex_to_rgb(value)
    if (value[0]*0.299 + value[1]*0.587 + value[2]*0.114) > intensity:
        return 'black'
    return 'white'


def check_param(master, name, cnf={}, **kw):
    "Internal function. Used to validate options of the widget."
    kw = _cnfmerge((cnf, kw))
    cm = Check_Comman_Parameters(master)
    return cm(name, kw)


def delta(evt):
    """Modified delta to work with all platforms."""
    if evt.num == 5 or evt.delta < 0:
        return -1
    return 1


def hex_to_rgb(hx, hsl=False):
    """Converts a HEX code into RGB or HSL.
    Args:
        hx (str): Takes both short as well as long HEX codes.
        hsl (bool): Converts the given HEX code into HSL value if True.
    Return:
        Tuple of length 3 consisting of either int or float values.
    Raise:
        ValueError: If given value is not a valid HEX code."""
    if re.compile(r'#[a-fA-F0-9]{3}(?:[a-fA-F0-9]{3})?$').match(hx):
        div = 255.0 if hsl else 0
        if len(hx) <= 4:
            return tuple(int(hx[i]*2, 16) / div if div else
                         int(hx[i]*2, 16) for i in (1, 2, 3))
        return tuple(int(hx[i:i+2], 16) / div if div else
                     int(hx[i:i+2], 16) for i in (1, 3, 5))
    raise ValueError(f'"{hx}" is not a valid HEX code.')


def get_shade(color, shade, mode='auto'):
    """### Darken or Lighten a shade of color.
    #### Args:
    1. `color`: Give a color as either HEX or name of the color.
    2. `shade`: The amount of change required. Takes float \
                between 0.0 to 1.0 eg: shade=0.225.
    3. `mode`:
        - `'-'` for darker shade.
        - `'+'` for lighter shade.
        - `'auto-110'` automatically decide lighter or \
                       darker. where 110 is the intensity.

    return hexcode"""
    op = {'+': lambda x, y: float(x+y),
          '-': lambda x, y: float(x-y)}
    if isinstance(color, str):
        try:
            color = list(float(i/65535.0)
                         for i in tkinter._default_root.winfo_rgb(color))
        except AttributeError:
            # Raise AttributeError when running tests.
            color = list(colour.Color(color).get_rgb())
    if 'auto' in mode:
        intensity = (110.0 if len(mode) <= 4 else float(
            mode.split('-')[1])) / 1000.0
        color_intensity = float(
            color[0]*0.299 + color[1]*0.587 + color[2]*0.114)
        mode = '-' if color_intensity > intensity else '+'
        if color_intensity > intensity*2 or color_intensity < intensity/2:
            shade += shade
    if mode not in op:
        raise ValueError(
            'Invalid mode "{}", Takes only "-" or "+"'. format(mode))
    for index, c in enumerate((op[mode](c, shade) for c in color)):
        if c > 1.0:
            c = 1.0
        elif c < 0.0:
            c = 0.0
        color[index] = int(c*255.0)
    return '#%02x%02x%02x' % (color[0], color[1], color[2])


def gradient(iteration):
    """This function returns a list of HSL values
    of all the colors in an order."""

    ops = {'+': lambda c, step: min(1.0, c + step),
           '-': lambda c, step: max(0.0, c - step)}

    index = 0
    operation = '+'
    iteration = max(1, iteration-2)
    rgb, _list = [1.0, 0.0, 0.0], []
    combo = ((2, 1, 0), (2, 0, 1), (0, 2, 1), (0, 1, 2), (1, 0, 2), (1, 2, 0))
    step = float(len(combo)) / float(iteration)
    _list.append('#%02x%02x%02x' % (round(rgb[0]*255),
                                    round(rgb[1]*255), round(rgb[2]*255)))
    for i in range(iteration):
        if (rgb[combo[index][1]] == 1.0 and operation == '+') or \
           (rgb[combo[index][1]] == 0.0 and operation == '-'):
            operation = '-' if operation == '+' else '+'
            index += 1
        rgb[combo[index][1]] = ops[operation](rgb[combo[index][1]], step)
        _list.append('#%02x%02x%02x' % (round(rgb[0]*255),
                                        round(rgb[1]*255), round(rgb[2]*255)))
    _list.append('#%02x%02x%02x' % (round(1.0*255),
                                    round(0.0*255), round(0.0*255)))
    return _list


__all__ = [
    'check_appearance',
    'check_light_dark',
    'delta',
    'get_shade',
    'gradient'
]



#                       Copyright 2021 Saad Mairaj
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tkmacosx.basewidgets.radiobutton_base import RadiobuttonBase


class Radiobutton(RadiobuttonBase):
    """Radiobutton widget which shows only one of several buttons in on-state."""

    def __init__(self, master=None, cnf={}, **kw):
        """Construct a radiobutton widget with the parent MASTER.

        Valid resource names: activebackground, activeforeground, anchor,
        background, bd, bg, bitmap, borderwidth, command, cursor,
        disabledforeground, fg, font, foreground, height,
        highlightbackground, highlightcolor, highlightthickness, image,
        indicatoron, justify, padx, pady, relief, selectcolor, selectimage,
        state, takefocus, text, textvariable, underline, value, variable,
        width, wraplength."""
        RadiobuttonBase.__init__(self, master=master, cnf=cnf, **kw)



#                       Copyright 2021 Saad Mairaj
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import tkinter
from tkmacosx.utils import (_cnfmerge, _bind, delta, check_param)


SFRAME_PROPERTIES = [
    'avoidmousewheel', 'autohidescrollbar', 'autohidescrollbardelay',
    'background', 'borderwidth', 'canvas', 'class', 'colormap', 'container',
    'cursor', 'height', 'highlightbackground', 'highlightcolor',
    'highlightthickness', 'mousewheel', 'padx', 'pady', 'relief',
    'scrollbar', 'scrollbarwidth', 'takefocus', 'visual', 'width',
]


SFRAME_FEATURES = (
    'scrollbarwidth', 'mousewheel', 'autohidescrollbar', 'avoidmousewheel',
    'canvas', 'scrollbar', 'autohidescrollbardelay'
)


class _sframe_functions:

    def _auto_hide_scrollbar(self, evt=None):
        "Internal function"
        if not self.cnf['autohidescrollbar']:
            return
        self.after_cancel(self._after_ids.get(1, ' '))
        if ((isinstance(evt, tkinter.Event) and evt.x >= self['scrollbar'].winfo_x())
                or evt == 'show'):
            self['scrollbar'].place(relx=1, rely=0, anchor='ne', relheight=1)
            self['scrollbar'].lift()
        else:
            self._after_ids[1] = self.after(
                self['autohidescrollbardelay'], self['scrollbar'].place_forget)

    def _avoid_mousewheel(self, widgets):
        """Internal function.\n
        Use this to have multiple scrollable 
        widgets inside of SFrame."""

        def set_widget(wid):
            """Internal function.\n
            Binds <Enter> and <Leave> to the widget 
            to enable/disable mousewheel scrolling."""
            binds = [{'className': 'mw_state_sframe', 'sequence':
                      '<Leave>', 'func': lambda _: self._mouse_scrolling(True)}]
            if not isinstance(wid, SFrameBase):
                binds.append({'className': 'mw_state_sframe', 'sequence':
                              '<Enter>', 'func': lambda _: self._mouse_scrolling(False)})
            _bind(wid, *binds)

        if isinstance(widgets, (list, tuple)):
            for widget in widgets:
                set_widget(widget)
        else:
            set_widget(widgets)

    def _configure_height(self, evt):
        """Internal function."""
        width = self.cnf['canvas'].winfo_width(
        )-self.cnf['scrollbar'].winfo_width()
        self.cnf['canvas'].itemconfig('window', width=width)

    def _configure_window(self, evt):
        """Internal function."""
        # this will update the position of scrollbar when scrolled from mousewheel.
        # fixes some bugs
        # makes scrolling more smoother
        self.after_cancel(self._after_ids.get(0, ' '))
        self._after_ids[0] = self.after(
            1, lambda: self.cnf['canvas'].configure(
                scrollregion=self.cnf['canvas'].bbox('all')))

    def _geometryManager(self):
        """Internal function."""
        # Use set to support the following in both python 2 and python 3
        geo_methods = [m for m in (set(tkinter.Pack.__dict__) | set(tkinter.Grid.__dict__) |
                                   set(tkinter.Place.__dict__)) if m not in tkinter.Frame.__dict__]
        for m in geo_methods:
            if m[0] != '_' and 'config' not in m:
                setattr(self, m, getattr(self.cnf['canvas'], m))

    def _mouse_scrolling(self, state):
        """Internal function."""
        def enable_mousewheel(evt=None):
            """Internal function."""
            self.bind_all('<MouseWheel>', self._on_mouse_scroll)

        def disable_mousewheel(evt=None):
            """Internal function."""
            self.unbind_all('<MouseWheel>')

        if state:
            _bind(self,
                  {'className': 'mousewheel_state', 'sequence':
                   '<Enter>', 'func': enable_mousewheel},
                  {'className': 'mousewheel_state', 'sequence':
                   '<Leave>', 'func': disable_mousewheel})
            enable_mousewheel()
        else:
            _bind(self,
                  {'className': 'mousewheel_state', 'sequence': '<Enter>'},
                  {'className': 'mousewheel_state', 'sequence': '<Leave>'})
            disable_mousewheel()

    def _on_mouse_scroll(self, evt):
        """Internal function."""
        if self.winfo_height() < self.cnf['canvas'].winfo_height():
            return
        if evt.state == 0:
            delay = 1
            if self['autohidescrollbardelay'] < 1000:
                delay = 1000 - self['autohidescrollbardelay']
            self.cnf['canvas'].yview_scroll(-1*delta(evt), 'units')
            self.after_cancel(self._after_ids.get(2, ' '))
            self._auto_hide_scrollbar('show')
            self._after_ids[2] = self.after(delay, self._auto_hide_scrollbar)

    def _set_configure(self, cnf):
        "Internal function."
        kw_copy = check_param(self, 'sframe',
                              {k: cnf.pop(k) for k in SFRAME_FEATURES if k in cnf})
        self.cnf['scrollbar']['width'] = kw_copy.pop(
            'scrollbarwidth', self.cnf['scrollbar']['width'])
        for key in kw_copy.copy():
            if key == 'avoidmousewheel':
                l = list(self.cnf[key])
                if isinstance(kw_copy[key], (tuple, list)):
                    l.extend(kw_copy.pop(key))
                else:
                    l.append(kw_copy.pop(key))
                self.cnf[key] = tuple(set(l))
            else:
                self.cnf[key] = kw_copy.pop(key, self.cnf.get(key))

        self.cnf['canvas']['width'] = cnf.pop(
            'width', self.cnf['canvas']['width'])
        self.cnf['canvas']['height'] = cnf.pop(
            'height', self.cnf['canvas']['height'])

        self._mouse_scrolling(self.cnf['mousewheel'])
        self._avoid_mousewheel(self.cnf['avoidmousewheel'])
        return cnf


class SFrameBase(tkinter.Frame, _sframe_functions):
    """Base Class for SFrame."""

    def __init__(self, master=None, cnf={}, **kw):
        kw = _cnfmerge((cnf, kw))
        kw['class_'] = kw.get('class_', 'Sframe')
        self._after_ids = {}
        self._over_scrollbar = False
        cnf = {}
        cnf['scrollbarwidth'] = kw.pop('scrollbarwidth', 10)
        cnf['mousewheel'] = kw.pop('mousewheel', True)
        cnf['avoidmousewheel'] = kw.pop('avoidmousewheel', ())
        cnf['autohidescrollbar'] = kw.pop('autohidescrollbar', False)
        cnf['autohidescrollbardelay'] = kw.pop('autohidescrollbardelay', 1000)
        cnf['canvas'] = kw.pop(
            'canvas', tkinter.Canvas(master=master, highlightthickness=0,
                                     width=kw.pop('width', 250), height=kw.pop('height', 250)))
        cnf['scrollbar'] = kw.pop(
            'scrollbar', tkinter.Scrollbar(cnf['canvas'], orient='vertical',
                                           width=cnf.get('scrollbarwidth')))
        tkinter.Frame.__init__(self, cnf['canvas'], **kw)
        self.cnf = check_param(self, 'sframe', **cnf)
        self.cnf['canvas']['bg'] = self['bg']

        if not self.cnf['autohidescrollbar']:
            self.cnf['scrollbar'].place(
                relx=1, rely=0, anchor='ne', relheight=1)

        self.cnf['scrollbar'].configure(command=self.cnf['canvas'].yview)
        self.cnf['canvas'].configure(yscrollcommand=self.cnf['scrollbar'].set)
        self.cnf['canvas'].create_window(
            0, 0, anchor='nw', tags="window", window=self,
            width=self.cnf['canvas'].winfo_reqwidth()-self.cnf['scrollbar'].winfo_reqwidth())
        self.cnf['canvas'].bind("<Configure>", self._configure_height, add="+")

        _bind(self, className='configure',
              sequence='<Configure>', func=self._configure_window)
        _bind(self, className='auto_hide_motion',
              sequence='<Motion>', func=self._auto_hide_scrollbar)
        _bind(
            self['scrollbar'], className='auto_hide_motion', sequence='<Enter>',
            func=lambda _: self._auto_hide_scrollbar('show'))
        _bind(
            self['scrollbar'], className='auto_hide_motion', sequence='<Leave>',
            func=self._auto_hide_scrollbar)

        self._mouse_scrolling(self.cnf['mousewheel'])
        self._avoid_mousewheel(self.cnf.get('avoidmousewheel'))
        self._geometryManager()

    def _configure(self, cmd, cnf=None, kw=None):
        cnf = self._set_configure(_cnfmerge((cnf, kw)))
        _return = tkinter.Frame._configure(self, cmd, None, cnf)
        if kw.get('bg', kw.get('background')):
            self.cnf['canvas']['bg'] = self['bg']
        if isinstance(_return, dict):
            _return.update(self.cnf)
            for k in list(_return):
                if k not in SFRAME_PROPERTIES:
                    _return.pop(k)
        return _return

    def keys(self):
        """Return a list of all resource names of this widget."""
        return SFRAME_PROPERTIES

    def cget(self, key):
        """Return the resource value for a KEY given as string."""
        if key in ('width', 'height'):
            return int(self.cnf['canvas'][key])
        if key == 'avoidmousewheel' and len(self.cnf.get(key)) == 1:
            return self.cnf[key][0]
        if key == 'scrollbarwidth':
            return int(self['scrollbar'].cget('width'))
        if key in SFRAME_FEATURES:
            return self.cnf.get(key)
        return tkinter.Frame.cget(self, key)
    __getitem__ = cget



import tkinter
from tkinter import TclError, Variable


UNITS = {
    'c': 72 / 2.54,     # centimeters
    'i': 72,            # inches
    'm': 72 / 25.4,     # millimeters
    'p': 1,             # points
}


def pixels_conv(value):
    return float(value[:-1]) * UNITS[value[-1:]]


class widget_options:

    button = {
        'boolean': ('borderless', ),
        'color': ('activebackground', 'activeforeground', 'bordercolor',
                  'disabledbackground', 'disabledforeground', 'foreground', 'fg',
                  'overforeground', 'overbackground', 'focuscolor', 'highlightbackground'),
        'enum': ('anchor', 'compound', 'state', 'justify', 'overrelief'),
        'integer': ('focusthickness', 'repeatinterval', 'repeatdelay', 'underline', 'radius'),
        'pixel': ('highlightthickness', 'padx', 'pady', 'width', 'height'),
        'not_exist': ('activeimage', 'image', 'font'),
    }

    colorscale = {
        'boolean': ('showinfo', 'mousewheel'),
        'enum': ('orient', 'value'),
        'integer': ('showinfodelay', ),
    }

    marquee = {
        'color': ('fg', 'foreground', 'disabledforeground'),
        'enum': ('justify', ),
        'integer': ('end_delay', 'fps', 'initial_delay', 'smoothness', 'left_margin'),
        'not_exist': ('font', ),
    }

    sframe = {
        'boolean': ('autohidescrollbar', 'mousewheel'),
        'integer': ('autohidescrollbardelay', ),
        'widget': ('canvas', 'scrollbar'),
    }

    @classmethod
    def get_options(cls, widget, name):
        return getattr(cls, widget, None)[name]

    @classmethod
    def get_keys(cls, widget):
        return list(getattr(cls, widget, None).keys())


class Check_Comman_Parameters:

    enum_para = {
        'anchor': ('n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw', 'center'),
        'compound': ('bottom', 'center', 'left', 'none', 'right', 'top'),
        'state': ('active', 'disabled', 'normal', 'pressed'),
        'justify': ('left', 'right', 'center'),
        'overrelief': ('flat', 'groove', 'raised', 'ridge', 'solid', 'sunken', ''),
        'orient': ('horizontal', 'vertical'),
        'value': ('rgb', 'hex'),
    }

    def __init__(self, master):
        self.master = master

    def __call__(self, widget, kw):
        for i in widget_options.get_keys(widget):
            fn = getattr(self, i)
            fn(widget, kw)
        return kw

    def boolean(self, widget, kw):
        for i in widget_options.get_options(widget, self.boolean.__name__):
            if i in kw:
                if kw[i] in (False, 0, 'false', 'no', 'off'):
                    kw[i] = 0
                elif kw[i] in (True, 1, 'true', 'yes', 'on'):
                    kw[i] = 1
                else:
                    raise TclError(
                        'expected boolean value but got "%s"' % kw[i])

    def color(self, widget, kw):
        for i in widget_options.get_options(widget, self.color.__name__):
            if i in kw and kw[i] != '':
                if i == 'activebackground' and isinstance(kw[i], (list, tuple)):
                    for c in kw[i]:
                        self.master.winfo_rgb(c)
                elif not isinstance(kw[i], Variable):
                    self.master.winfo_rgb(kw[i])

    def enum(self, widget, kw):
        op_list = widget_options.get_options(widget, self.enum.__name__)
        for k, v in self.enum_para.items():
            if (k in op_list and k in kw
                and kw[k] not in v
                and not (k == 'compound'
                                 and str(kw[k]).lower() in 'none'
                                 and kw[k] != '')
                ):
                err = 'ambiguous' if kw[k] == '' else 'bad'
                name = 'justification' if k == 'justify' else k
                if 'overrelief' == k:
                    v = v[:-1]
                    name = 'relief'
                raise TclError(
                    '%s %s "%s": must be %s%s or %s' % (
                        err, name, kw[k], ', '.join(v[:-1]),
                        ',' if len(v) > 2 else '', v[-1]))

    def integer(self, widget, kw):
        for i in widget_options.get_options(widget, self.integer.__name__):
            if i in kw:
                if kw[i] == '' or '.' in str(kw[i]) or not str(kw[i]).lstrip("-").isnumeric():
                    raise TclError('expected integer but got "%s"' % kw[i])

    def not_exist(self, widget, kw):
        for i in widget_options.get_options(widget, self.not_exist.__name__):
            image_cond = ('image' in i and i in kw and
                          isinstance(kw[i], str) and kw[i] != '')
            font_cond = 'font' in kw and kw['font'] == ''
            if i in kw and (image_cond or font_cond):
                name = 'image' if i == 'activeimage' else i
                raise TclError('%s "%s" doesn\'t exist' % (name, kw[i]))

    def pixel(self, widget, kw):
        for i in widget_options.get_options(widget, self.pixel.__name__):
            if i in kw:
                if '.' in str(kw[i]):
                    kw[i] = round(float(kw[i]))
                if isinstance(kw[i], str):
                    if kw[i][-1] in UNITS and kw[i][:-1].isnumeric():
                        kw[i] = round(pixels_conv(kw[i]))
                if not isinstance(kw[i], (int, float)):
                    raise TclError('bad screen distance "%s"' % kw[i])

    def widget(self, widget, kw):
        for i in widget_options.get_options(widget, self.widget.__name__):
            if i in kw:
                name = i.capitalize()
                if not isinstance(kw[i], getattr(tkinter, name)):
                    raise TclError(
                        'expected tkinter.%s instance but got "%s"' % (name, kw[i]))



#                       Copyright 2021 Saad Mairaj
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import random
import tkinter as tk

from tkinter import font
from tkinter import ttk
from tkinter.colorchooser import askcolor

from tkmacosx import SFrame, Button, Colorscale, ColorVar
from tkmacosx.utils import check_light_dark


def grid(root, row, column):
    "Defines rows and columns if grid method is used"
    for y in range(column):
        tk.Grid.columnconfigure(root, y, weight=1)
    for x in range(row):
        tk.Grid.rowconfigure(root, x, weight=1)


def sort_random_values(fn):
    def wrapper(*args, **kwargs):
        value = fn(*args, **kwargs)
        if isinstance(value, (list, tuple)) and len(value) == 1:
            return value[0]
        return value
    return wrapper


def choices(seq, k):
    "Random choices."
    return tuple(random.choice(seq) for i in range(k))


@sort_random_values
def get_random_colors(k=1):
    return ["#%06x" % random.randint(0, 0xFFFFFF) for n in range(k)]


@sort_random_values
def get_random_font(k=1):
    def family():
        return random.choice(font.families())

    def weight():
        return random.choice(('bold', 'normal'))

    def slant():
        return random.choice(('italic', 'roman'))

    def underline():
        return random.choice((True, False))

    overstrike = underline
    return tuple(font.Font(family=family(), weight=weight(),
                           slant=slant(), underline=underline(),
                           overstrike=overstrike(), size=30,) for i in range(k))


@sort_random_values
def get_random_size(k=1, start=10, end=200):
    return tuple(choices([i for i in range(start, end)], k=k))


@sort_random_values
def get_random_relief(k=1):
    return choices(('flat', 'groove', 'raised',
                    'ridge', 'solid', 'sunken'), k=k)


class Sample(tk.Tk):
    def __init__(self):
        tk.Tk.__init__(self)
        self.resizable(0, 0)
        self.geometry('420x700+300+100')
        self.title('Mac OSX Button Testing')
        self.wm_attributes('-modified', 1)
        self.main_color = ColorVar(value='#ffe6f4')
        self['bg'] = self.main_color
        grid(self, 25, 5)
        self.L1 = tk.Label(self, text='Comparison',
                           bg=self.main_color, font=('', 18, 'bold'))
        self.L1.grid(row=0, column=0, columnspan=5, sticky='nsew')
        Button(self, text='Hello').grid(row=1, column=1, sticky='s')
        ttk.Button(self, text='Hello').grid(row=1, column=3, sticky='s')
        tk.Button(self, text='Hello').grid(row=1, column=2, sticky='s')
        tk.Label(self, bg=self.main_color, font=('', 10),
                 text='(Mac OSX)').grid(row=2, column=1, sticky='n',)
        tk.Label(self, bg=self.main_color, font=('', 10),
                 text='(ttk themed)').grid(row=2, column=3, sticky='n')
        tk.Label(self, bg=self.main_color, font=('', 10),
                 text='(Default)').grid(row=2, column=2, sticky='n')
        ttk.Separator(self, orient='vertical').grid(
            row=3, column=0, columnspan=5, sticky='ew')

        # ------------ Seperator -------------

        # ------------ Demonstration ------------

        self.sfr = SFrame(self, bg=self.main_color)
        self.sfr.grid(rowspan=27, columnspan=5, sticky='nsew')
        for i in range(5):
            self.sfr.grid_columnconfigure(i, weight=1)
        self.L2 = tk.Label(self.sfr, text='Demonstration',
                           bg=self.main_color, font=('', 20, 'bold'))
        self.L2.grid(row=1, column=0, columnspan=5,
                     sticky='new', pady=(20, 10))

        # ------------ Active Color ------------

        self.L3 = tk.Label(self.sfr, text='1. Change Active color', bg=self.main_color,
                           font=('', 15, 'bold'))
        self.L3.grid(row=2, column=0, columnspan=5, sticky='nsew', pady=10)
        self.L4 = tk.Label(self.sfr, text='The active color can be changed to any gradient color.',
                           bg=self.main_color, font=('', 10))
        self.L4.grid(row=3, column=0, columnspan=5, sticky='new')
        self.B1 = Button(self.sfr, text='Press Me', pady=20)
        self.B1.grid(row=4, column=0, columnspan=5, pady=20)
        self.C1 = tk.StringVar(value='Select')
        self.L5 = tk.Label(self.sfr, text='From',
                           bg=self.main_color, font=('', 12))
        self.L5.grid(row=5, column=1, sticky='ew')

        self.L6 = tk.Label(self.sfr, text='To',
                           bg=self.main_color, font=('', 12))
        self.L6.grid(row=5, column=3, sticky='ew')
        
        button_container1 = tk.Frame(self.sfr, bg=self.main_color)
        button_container1.grid(row=6, column=0, columnspan=5,
                              sticky='ew', pady=10, padx=10)
        button_container1.columnconfigure(0, weight=1)
        button_container1.columnconfigure(1, weight=1)
        self.Om1_b = Button(button_container1, text='Change Color',
                         bg='#d0c0ea', borderless=1)
        self.Om1_b.grid(row=0, column=0, padx=(0, 2), sticky='ew')
        self.Om2_b = Button(button_container1, text='Change Color',
                         bg="#d0c0ea", borderless=1)
        self.Om2_b.grid(row=0, column=1, sticky='ew')
        self.Om1_b['command'] = lambda: self.change_active_color(1)
        self.Om2_b['command'] = lambda: self.change_active_color(2)

        # ------------ Background Color ------------

        self.L7 = tk.Label(self.sfr, text='2. Change Background color', bg=self.main_color,
                           font=('', 15, 'bold'))
        self.L7.grid(row=7, column=0, columnspan=5,
                     sticky='nsew', pady=(50, 0))
        self.L8 = tk.Label(self.sfr, text='Click on the button to choose the color.',
                           bg=self.main_color, font=('', 10))
        self.L8.grid(row=8, column=0, columnspan=5, sticky='new', pady=10)

        self.B2 = Button(self.sfr, text='Color me',
                         font=('', 30,), pady=10)
        self.B2.grid(row=9, column=0, columnspan=5, sticky='', pady=20)

        button_container2 = tk.Frame(self.sfr, bg=self.main_color)
        button_container2.grid(row=10, column=0, columnspan=5,
                              sticky='ew', pady=10, padx=10)
        button_container2.columnconfigure(0, weight=1)
        button_container2.columnconfigure(1, weight=1)
        self.B3 = Button(button_container2, text='Change Background Color',
                         bg='#d0c0ea', borderless=1)
        self.B3['command'] = lambda: self.B2.config(bg=askcolor()[1])
        self.B3.grid(row=0, column=0, padx=(0, 2))
        self.B4 = Button(button_container2, text='Change Foreground Color',
                         bg="#d0c0ea", borderless=1)
        self.B4['command'] = lambda: self.B2.config(fg=askcolor()[1])
        self.B4.grid(row=0, column=1)

        # ------------ Borderless ------------

        self.L9 = tk.Label(self.sfr, text='3. Switch Between Borderless', bg=self.main_color,
                           font=('', 15, 'bold'))
        self.L9.grid(row=11, column=0, columnspan=5,
                     sticky='sew', pady=(50, 0))
        self.L10 = tk.Label(self.sfr, text="""
    In borderless it will blend with its parent widget background color.
    Give parameter `borderless = True / False` to use it.""", bg=self.main_color, font=('', 10))
        self.L10.grid(row=12, column=0, columnspan=5, sticky='new')

        self.B5 = Button(self.sfr, text='No Borders', borderless=1, height=40,
                         bg='#212F3D', fg='white', activebackground=("#EAECEE", "#212F3D"))
        self.B5.grid(row=13, columnspan=5, pady=(20, 5))

        self.B6 = Button(self.sfr, text='No Borders', borderless=1, height=40,
                         bg='#F7DC6F', fg='#21618C', activebackground=('#B3B6B7', '#58D68D'))
        self.B6.grid(row=14, columnspan=5, pady=(0, 20))
        self.var1 = tk.BooleanVar(value=True)
        self.CB1 = tk.Checkbutton(self.sfr, text='Toggle Borderless', variable=self.var1,
                                  bg=self.main_color, command=self.change_borderless_state)
        self.CB1.grid(row=15, columnspan=5, pady=(0, 10))

        # ------------ Bordercolor ------------

        self.L11 = tk.Label(self.sfr, text='4. Change Bordercolor', bg=self.main_color,
                            font=('', 15, 'bold'))
        self.L11.grid(row=16, column=0, columnspan=5,
                      sticky='sew', pady=(50, 0))
        self.L12 = tk.Label(self.sfr, text="Change Bordercolor of the button\nNote: if borderless=True, then the bordercolor won't work.",
                            bg=self.main_color, font=('', 10))
        self.L12.grid(row=17, column=0, columnspan=5, sticky='new')

        self.B7 = Button(self.sfr, text='Button', pady=10,
                         padx=5, font=('Zapfino', 12, 'bold'))
        self.B7.grid(row=18, columnspan=5, pady=30)

        self.CS1 = Colorscale(self.sfr, value='hex', mousewheel=1,
                              command=lambda e: self.B7.config(bordercolor=e))
        self.CS1.grid(row=19, columnspan=5, sticky='ew', padx=10)

        self.CS2 = Colorscale(self.sfr, value='hex', mousewheel=1,
                              gradient=('#FCF6F5', '#990011'),
                              command=lambda e: self.B7.config(bg=e))
        self.CS2.grid(row=20, columnspan=5, sticky='ew', padx=10, pady=5)

        self.CS3 = Colorscale(self.sfr, value='hex', mousewheel=1,
                              gradient=('green', 'yellow'),
                              command=lambda e: self.B7.config(fg=e))
        self.CS3.grid(row=21, columnspan=5, sticky='ew', padx=10)

        self.CS4 = Colorscale(self.sfr, value='hex', mousewheel=1,
                              gradient=('pink', 'purple'),
                              command=lambda e: self.B7.config(overforeground=e))
        self.CS4.grid(row=22, columnspan=5, sticky='ew', padx=10, pady=5)

        # ------------ Random button styling ------------
        self.L11 = tk.Label(self.sfr, text='5. Button Styling', bg=self.main_color,
                            font=('', 15, 'bold'))
        self.L11.grid(row=23, column=0, columnspan=5,
                      sticky='sew', pady=(50, 0))
        self.L12 = tk.Label(self.sfr, text="Press the button to ramdomise the style of the button.",
                            bg=self.main_color, font=('', 10))
        self.L12.grid(row=24, column=0, columnspan=5, sticky='new')

        self.B10 = Button(self.sfr, text='Button', borderless=1)
        self.B10.grid(row=25, columnspan=5, pady=20)

        self.B11 = Button(self.sfr, text='Change Style', borderless=1, fg='#21618C',
                          activebackground=('#B3B6B7', '#58D68D'), command=self.change_button_style)
        self.B11.grid(row=26, columnspan=5, ipady=5)

        self.button_clicks = 1
        self.Text1 = tk.Text(self.sfr, background=self.main_color, highlightthickness=0,
                             relief='sunken', height=20, bd=2, padx=10)
        self.Text1.grid(row=27, columnspan=5, pady=20, padx=20)
        self.sfr._avoid_mousewheel(
            (self.Text1, self.CS1, self.CS2, self.CS3, self.CS4))
        self.change_button_style()
        self.update_idletasks()

    def change_active_color(self, num):
        hexcode = askcolor()[1]
        if num == 1:
            self.Om1_b['bg'] = self.Om1_b['text'] = hexcode
            self.Om1_b['fg'] = check_light_dark(hexcode)
        else:
            self.Om2_b['bg'] = self.Om2_b['text'] = hexcode
            self.Om2_b['fg'] = check_light_dark(hexcode)
        self.B1['activebackground'] = (self.Om1_b['bg'], self.Om2_b['bg'])

    def change_borderless_state(self):
        if self.var1.get():
            self.B5['borderless'] = 1
            self.B6['borderless'] = 1
        else:
            self.B5['borderless'] = 0
            self.B6['borderless'] = 0

    def change_button_style(self):
        cnf = dict(
            foreground=get_random_colors(),
            background=get_random_colors(),
            activebackground=get_random_colors(2),
            activeforeground=get_random_colors(),
            overrelief=get_random_relief(),
            relief=get_random_relief(),
            highlightthickness=get_random_size(1, 1, 5),
            borderwidth=get_random_size(1, 1, 5),
            font=get_random_font(),
            highlightbackground=get_random_colors(),
            focuscolor=get_random_colors(),
            overbackground=get_random_colors(),
            overforeground=get_random_colors(),
        )
        self.B10.config(**cnf)

        self.Text1.insert('end', '\n\nBUTTON STYLE %s\n\n' %
                          self.button_clicks)
        for k, v in cnf.items():
            if k == 'font':
                self.Text1.insert('end', 'font:\n')
                for i, j in v.config().items():
                    self.Text1.insert('end', '\t%s\t\t= %s\n' % (i, j))
            else:
                self.Text1.insert('end', '%s\t\t\t= %s\n' % (k, v))
        self.Text1.see('end')
        self.Text1.insert('end', '\n\n'+'-'*40)
        self.button_clicks += 1
        self.sfr['canvas'].yview_moveto('1.0')
        self.update_idletasks()


def main():
    "Demonstration of tkmacosx."
    Sample().mainloop()


#  Testing Demo
if __name__ == "__main__":
    main()



import _tkinter
import functools
import re
import tkinter
import unittest
import os
import sys


# dummy value to run tests without test.support
verbose = 1   # Flag set to 0 by regrtest.py file in test folder.

# TEST_HOME_DIR refers to the top level directory of the "test" package
# that contains Python's regression test suite
TEST_SUPPORT_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_HOME_DIR = os.path.dirname(TEST_SUPPORT_DIR)


# Class of cpython/Lib/test/support/__init__.py
class ResourceDenied(unittest.SkipTest):
    """Test skipped because it requested a disallowed resource.
    This is raised when a test calls requires() for a resource that
    has not be enabled.  It is used to distinguish between expected
    and unexpected skips.
    """


def findfile(filename, subdir=None):
    """Try to find a file on sys.path or in the test directory.  If it is not
    found the argument passed to the function is returned (this does not
    necessarily signal failure; could still be the legitimate path).
    Setting *subdir* indicates a relative path to use to find the file
    rather than looking directly in the path directories.
    """
    if os.path.isabs(filename):
        return filename
    if subdir is not None:
        filename = os.path.join(subdir, filename)
    path = [TEST_HOME_DIR] + sys.path
    for dn in path:
        fn = os.path.join(dn, filename)
        if os.path.exists(fn):
            return fn
    return filename


class AbstractTkTest:

    @classmethod
    def setUpClass(cls):
        cls._old_support_default_root = tkinter._support_default_root
        destroy_default_root()
        tkinter.NoDefaultRoot()
        cls.root = tkinter.Tk()
        cls.wantobjects = cls.root.wantobjects()
        # De-maximize main window.
        # Some window managers can maximize new windows.
        cls.root.wm_state('normal')
        try:
            cls.root.wm_attributes('-zoomed', False)
        except tkinter.TclError:
            pass

    @classmethod
    def tearDownClass(cls):
        cls.root.update_idletasks()
        cls.root.destroy()
        del cls.root
        tkinter._default_root = None
        tkinter._support_default_root = cls._old_support_default_root

    def setUp(self):
        self.root.deiconify()

    def tearDown(self):
        for w in self.root.winfo_children():
            w.destroy()
        self.root.withdraw()


def destroy_default_root():
    if getattr(tkinter, '_default_root', None):
        tkinter._default_root.update_idletasks()
        tkinter._default_root.destroy()
        tkinter._default_root = None


def simulate_mouse_click(widget, x, y):
    """Generate proper events to click at the x, y position (tries to act
    like an X server)."""
    widget.event_generate('<Enter>', x=0, y=0)
    widget.event_generate('<Motion>', x=x, y=y)
    widget.event_generate('<ButtonPress-1>', x=x, y=y)
    widget.event_generate('<ButtonRelease-1>', x=x, y=y)


tcl_version = tuple(map(int, _tkinter.TCL_VERSION.split('.')))


def requires_tcl(*version):
    if len(version) <= 2:
        return unittest.skipUnless(
            tcl_version >= version,
            'requires Tcl version >= ' + '.'.join(
                map(str, version)))

    def deco(test):
        @functools.wraps(test)
        def newtest(self):
            if get_tk_patchlevel() < version:
                self.skipTest(
                    'requires Tcl version >= ' + '.'.join(
                        map(str, version)))
            test(self)
        return newtest
    return deco


_tk_patchlevel = None


def get_tk_patchlevel():
    global _tk_patchlevel
    if _tk_patchlevel is None:
        tcl = tkinter.Tcl()
        patchlevel = tcl.call('info', 'patchlevel')
        m = re.fullmatch(r'(\d+)\.(\d+)([ab.])(\d+)', patchlevel)
        major, minor, releaselevel, serial = m.groups()
        major, minor, serial = int(major), int(minor), int(serial)
        releaselevel = {'a': 'alpha', 'b': 'beta', '.': 'final'}[releaselevel]
        if releaselevel == 'final':
            _tk_patchlevel = major, minor, serial, releaselevel, 0
        else:
            _tk_patchlevel = major, minor, 0, releaselevel, serial
    return _tk_patchlevel


def tcl_obj_eq(actual, expected):
    if actual == expected:
        return True
    if isinstance(actual, _tkinter.Tcl_Obj):
        if isinstance(expected, str):
            return str(actual) == expected
    if isinstance(actual, tuple):
        if isinstance(expected, tuple):
            return (len(actual) == len(expected) and
                    all(tcl_obj_eq(act, exp)
                        for act, exp in zip(actual, expected)))
    return False
